/**************************************************
 *
  File: bse-agent.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  NICE implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "nicenode.h"
#include <scheduler.h>
#include "app-packet.h"
#include "bse-agent.h"
#include "setpartition.h"
#include "common.h"

#define LOG_BSE_START
#define LOG_BSE_STOP

#undef LOG_BSE_SEND
#undef LOG_BSE_RECV
#undef LOG_BSE_JUNK
#undef LOG_BSE_PARTITION

#define LOG_BSE_CLUSTER_CHANGE
#undef LOG_CLUSTER_INFO_DETAILED

#define LOG_DATA_PACKET_SEND
#define LOG_DATA_PACKET_RECV

void bseAgent::init (int Id, int Index, Node *N) {

  Agent::init(Id,Index,N);
  t = AGENT_APPLICATION_BSE;
  top_layer = NULL;
  return;
}

void bseAgent::start (void) {

  Agent::start();

#ifdef LOG_BSE_START
  printf ("[bse %d %d ] At %8.4f start\n", id, n->id, Scheduler::Clock());
#endif // LOG_BSE_START

  reset_layer_info(0);
  log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
  display_top_layer_info();
#endif // LOG_BSE_CLUSTER_CHANGE

  return;
}

void bseAgent::stop (void) {

#ifdef LOG_BSE_STOP
  printf ("[bse %d %d ] At %8.4f stop\n", id, n->id, Scheduler::Clock());
#endif // LOG_BSE_STOP

  assert (started == true);
  if (top_layer != NULL)
    delete top_layer;

  Agent::stop();
  return;
}

void bseAgent::specific_rx_pkt_handler  (Packet *p) {

  if (p->t != PACKET_APP) {

     printf ("[Err] BSE received unknown packet type\n");
     return;
  }

  AppPacket *ap = (AppPacket *)p;

  rx_pkt_wrapper(ap);

  switch (ap->st) {
    
  case JOIN_QUERY :
    
    handle_join_query(ap);
    break;

  case CLUSTER_REFRESH :
    handle_cluster_refresh(ap);
    break;
    
  case CLUSTER_MERGE :
    handle_cluster_merge(ap);
    break;
    
  case JOIN_FORWARD :
    printf ("[Err] BSE received Join-forward, ignored\n");
    break;

  case PING_QUERY :
  case PING_RESPONSE :
    printf ("[Err] BSE received ping, ignored\n");
    break;

  case PACKET_DATA :
    handle_data_packet(ap);
    break;

  default :
    printf ("[Err] BSE received unknown packet subtype\n");
  }

  return;
}

void bseAgent::reset_layer_info (int top_lid) {

  if (top_layer != NULL) {
    top_layer->MakeEmpty();
    top_layer->lid = top_lid;
    // delete top_layer;
#ifdef LOG_BSE_CLUSTER_CHANGE
  printf ("[bse %d %d ] At %8.4f cluster-info : lid %d : deleted\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_CLUSTER_CHANGE
  }
  else
    top_layer = new LayerInfo(top_lid,this);

#ifdef LOG_BSE_JUNK
  printf ("[bse %d %d ] At %8.4f reset-layer : lid %d\n", id, n->id, Scheduler::Clock(), top_lid);
#endif // LOG_BSE_JUNK

  LayerAgentInfo * self = new LayerAgentInfo (id,n->id);
  self->dist = 0.0;
  top_layer->AddClusterRoot(self);
  top_layer->me_in_layer = true;
  top_layer->cr_chkt.CancelTimer();
  top_layer->cr_msgt.CancelTimer();
  top_layer->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);
  top_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  return;
}

void bseAgent::handle_join_query (AppPacket *p) {

  assert (p->st == JOIN_QUERY); // || (p->st == JOIN_FORWARD) );
  /* Join forward not accepted right now. To do so, changes needed:
     See BSE-JF change in coop-agent.cc
  */
  int qlid;
  AgentInfo src_ag;
  AgentInfo original_dst;
  double src_time;

  if (p->st == JOIN_QUERY) {
    qlid = p->u.joinq_p.q_lid;
    src_ag.agent_id = p->src_agent;
    src_ag.node_id = p->src;
    original_dst.agent_id = id;
    original_dst.node_id = n->id;
    src_time = p->u.joinq_p.send_time;
  }
  else {
    qlid = p->u.joinforward_p.q_lid;
    src_ag.agent_id = p->u.joinforward_p.src_ag.agent_id;
    src_ag.node_id = p->u.joinforward_p.src_ag.node_id;
    original_dst.agent_id = p->u.joinforward_p.original_dst.agent_id;
    original_dst.node_id = p->u.joinforward_p.original_dst.node_id;
    src_time = p->u.joinforward_p.send_time;
  }
    
#ifdef LOG_BSE_RECV
  printf ("[bse %d %d ] At %8.4f recvd-join-query : < [ %d %d ] > lid %d ( top lid %d )\n", id, n->id, Scheduler::Clock(), src_ag.agent_id, src_ag.node_id, qlid, top_layer->lid);
#endif // LOG_BSE_RECV

  if (top_layer->lid < qlid) {
    // Someone claims to need to join to a higher layer than I know of
    reset_layer_info(qlid);
  }

  bool do_attach = false;
  LayerAgentInfo * new_member = NULL;
  if (top_layer->lid == qlid) {
    do_attach = true;

    new_member = new LayerAgentInfo(src_ag.agent_id, src_ag.node_id);

    bool is_new_member = top_layer->AddClusterMember(new_member);
    if (is_new_member == false) {
      delete new_member;
      new_member = top_layer->FindClusterMember(src_ag.agent_id);
      new_member->refresh = true;
    }
    else {
      log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
      display_top_layer_info();
#endif // LOG_BSE_CLUSTER_CHANGE
    }
    new_member->dist = Scheduler::Clock() - src_time;

    //if (is_new_member == false)
    //  return;
  }
  /* Create response packet */
  AppPacket *resp_p = new AppPacket(JOIN_RESPONSE);
  put_layer_cluster_info_into_packet(top_layer,resp_p);
  resp_p->u.joinresp_p.accept = true;
  resp_p->u.joinresp_p.exp_src.agent_id = original_dst.agent_id;
  resp_p->u.joinresp_p.exp_src.node_id = original_dst.node_id;
  send_pkt_wrapper(resp_p,src_ag.agent_id, src_ag.node_id);

  if (do_attach == true) {
#ifdef LOG_BSE_SEND
    printf ("\n[bse %d %d ] At %8.4f : new-member < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), new_member->ag.agent_id, new_member->ag.node_id);
#endif // LOG_BSE_SEND
    top_layer->cr_msgt.CancelTimer();
    send_top_layer_cluster_refresh(new_member);
    top_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  }

  return;
}

// Accept the merge only if this is the only member in the top layer
// or if there is no member
void bseAgent::handle_cluster_merge (AppPacket * ap) {

  if ( top_layer->lid != ap->u.clustermerge_p.layer_id )
    return;

  if ( (top_layer->ag_list.GetSize() == 1) ||
       ( (top_layer->ag_list.GetSize() == 2) && (top_layer->ag_list.Locate(ap->src_agent) != NULL) ) ) {
#ifdef LOG_BSE_RECV
    printf ("[bse %d %d ] At %8.4f recvd-valid-merge : < [ %d %d ] > lid %d\n", id, n->id, Scheduler::Clock(), ap->src_agent, ap->src, ap->u.clustermerge_p.layer_id);
#endif // LOG_BSE_RECV
    reset_layer_info(ap->u.clustermerge_p.layer_id);
    top_layer->AddExtraMembersFromAltRootPacket(ap);

    log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
    display_top_layer_info();
#endif // LOG_BSE_CLUSTER_CHANGE
  }
  return;
}

void bseAgent::handle_cluster_refresh (AppPacket *ap) {

  if (top_layer->lid != ap->u.clusterrefresh_p.layer_id)
    return;
  void * pos = top_layer->ag_list.Locate(ap->src_agent);
  if (pos == NULL)
    return;

#ifdef LOG_BSE_RECV
  printf ("[bse %d %d ] At %8.4f recvd-valid-refresh : < [ %d %d ] > lid %d ", id, n->id, Scheduler::Clock(), ap->src_agent, ap->src, top_layer->lid);
#endif // LOG_BSE_RECV

  LayerAgentInfo *la_ag = top_layer->ag_list.GetAt(pos);
  if (ap->u.clusterrefresh_p.cluster_remove == true) {
#ifdef LOG_BSE_RECV
    printf ("*remove*\n");
#endif // LOG_BSE_RECV
    bool check = top_layer->DeleteClusterMember(la_ag);
    assert (check == true);
    delete la_ag;
    log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
    display_top_layer_info();
#endif // LOG_BSE_CLUSTER_CHANGE
  }
  else {
#ifdef LOG_BSE_RECV
    printf ("\n");
#endif // LOG_BSE_RECV
    la_ag->refresh_agent(ap);
  }

  return;
}

void bseAgent::handle_data_packet (AppPacket *ap) {

  assert (ap->st == PACKET_DATA);
  assert ( (ap->u.data_p.original_src.agent_id == ap->src_agent) && (ap->u.data_p.original_src.node_id == ap->src) );

#ifdef LOG_DATA_PACKET_RECV
  printf ("[bse %d %d ] At %8.4f fwd data-pkt : < [ %d %d ] > seq %d\n", id, n->id, Scheduler::Clock(), ap->u.data_p.original_src.agent_id, ap->u.data_p.original_src.node_id, ap->u.data_p.seq_no);
#endif // LOG_DATA_PACKET_RECV
  assert (top_layer != NULL);

  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       top_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = top_layer->ag_list.GetAt(pos);

    if ( (this_la->ag.agent_id == ap->u.data_p.original_src.agent_id) && (this_la->ag.node_id == ap->u.data_p.original_src.node_id) )
       continue;

    if (self_check(this_la) == false) {

       AppPacket *new_p = new AppPacket(PACKET_DATA);
       new_p->u.data_p.original_src.agent_id = ap->u.data_p.original_src.agent_id;
       new_p->u.data_p.original_src.node_id = ap->u.data_p.original_src.node_id;
       new_p->u.data_p.seq_no = ap->u.data_p.seq_no;
       new_p->u.data_p.lid = top_layer->lid;
#ifdef LOG_DATA_PACKET_SEND
       printf ("[bse %d %d ] At %8.4f send data-pkt : < [ %d %d ] > seq %d : to < [ %d %d ] > lid %d\n", id, n->id, Scheduler::Clock(), ap->u.data_p.original_src.agent_id, ap->u.data_p.original_src.node_id, ap->u.data_p.seq_no, this_la->ag.agent_id, this_la->ag.node_id, top_layer->lid);
#endif // LOG_DATA_PACKET_SEND
       send_pkt_wrapper(new_p,this_la->ag.agent_id, this_la->ag.node_id);
    }
  }
  return;
}
 

void bseAgent::handle_cluster_refresh_msg_timeout (void) {

  assert (top_layer != NULL);
#ifdef LOG_BSE_SEND
  printf ("\n[bse %d %d ] At %8.4f : refresh-timeout: lid %d\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_SEND

  send_top_layer_cluster_refresh(NULL);
  top_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  return;
}

void bseAgent::send_top_layer_cluster_refresh (LayerAgentInfo * ignore_agent) {

  assert (top_layer != NULL);

#ifdef LOG_BSE_SEND
  printf ("\n[bse %d %d ] At %8.4f (b-scr) : send-refresh : lid %d\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_SEND

  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       top_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = top_layer->ag_list.GetAt(pos);
    if (ignore_agent != NULL) {
      if ( (ignore_agent->ag.agent_id == la->ag.agent_id) && (ignore_agent->ag.node_id == la->ag.node_id) )
         continue;
    }

    if (self_check(la) == false) {

#ifdef LOG_BSE_SEND
  printf ("[bse %d %d ] . . (b-scr) : < [ %d %d ] >\n", id, n->id, la->ag.agent_id, la->ag.node_id);
#endif // LOG_BSE_SEND

      AppPacket *ap = new AppPacket (CLUSTER_REFRESH);
      put_layer_cluster_info_into_packet(top_layer,ap);
      ap->u.clusterrefresh_p.root_xfer = false;
      ap->u.clusterrefresh_p.is_root = true;
      send_pkt_wrapper(ap,la->ag.agent_id,la->ag.node_id);
    }
  }
#ifdef LOG_BSE_SEND
  printf ("\n");
#endif // LOG_BSE_SEND
  return;
}


void bseAgent::handle_cluster_refresh_check_timeout (void) {

  bool change = false;

  assert (top_layer != NULL);
#ifdef LOG_BSE_JUNK
      printf ("[bse %d %d ] At %8.4f (b-rc) : refresh-chk : lid %d\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_JUNK

  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       ) {

    LayerAgentInfo * la_ag = top_layer->ag_list.GetAt(pos);
    void * old_pos = pos;
    top_layer->ag_list.GetNext(pos);
    if (self_check(la_ag) == true)
      continue;
    if (la_ag->refresh == false) {
      /* Member is lost */
      top_layer->ag_list.RemoveAt(old_pos);
#ifdef LOG_BSE_JUNK
      printf ("[bse %d %d ] . . (b-rc) : < [ %d %d ] > *lost*\n", id, n->id, la_ag->ag.agent_id, la_ag->ag.node_id);
#endif // LOG_BSE_JUNK

      delete la_ag;
      change = true;
    }
    else {
      la_ag->refresh = false; // Resetting the refresh flag
#ifdef LOG_BSE_JUNK
      printf ("[bse %d %d ] . . (b-rc) : < [ %d %d ] >\n", id, n->id, la_ag->ag.agent_id, la_ag->ag.node_id);
#endif // LOG_BSE_JUNK
    }
  }
#ifdef LOG_BSE_JUNK
  printf ("\n");
#endif // LOG_BSE_JUNK

  assert (top_layer->ag_list.GetSize() >= 1);
  if (top_layer->ag_list.GetSize() == 1) { // I am the only one
    if (top_layer->lid > 0) {
#ifdef LOG_BSE_CLUSTER_CHANGE
  printf ("[bse %d %d ] At %8.4f cluster-info : lid %d : deleted\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_CLUSTER_CHANGE
      top_layer->lid --;
#ifdef LOG_BSE_JUNK
      printf ("[bse %d %d ] At %8.4f refresh-chk layer-decrement to lid %d\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_JUNK
      change = true;
    }
  }

  if (top_layer->ag_list.GetSize() >= (UPPER_3K+1)) {
    split_top_layer_using_two_partition();
    change = true;
  }

  if (change == true) {
    log_cluster_change_info ();
#ifdef LOG_BSE_CLUSTER_CHANGE
    display_top_layer_info();
#endif // LOG_BSE_CLUSTER_CHANGE
  }

  top_layer->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);
  return;
}

void bseAgent::split_top_layer_using_two_partition (void) {

#ifdef LOG_BSE_JUNK
  printf ("[bse %d %d ] At %8.4f cluster-split : lid %d\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_JUNK

  LayerAgentInfo ** ag_arr = top_layer->CreateLayerAgentInfoArray();
  assert (ag_arr != NULL);
  double * self_included_cost = create_cost_matrix(id,n->id,ag_arr,top_layer->ag_list.GetSize());
  // This function already updates the agent_arr by deleting this agent
  double * cost = delete_agent_from_cost_matrix(self_included_cost,ag_arr,top_layer->ag_list.GetSize(),id,n->id);
  free(self_included_cost);

  void * pos = top_layer->ag_list.Locate(id);
  assert (pos != NULL);
  LayerAgentInfo * self = top_layer->ag_list.GetAt(pos);
  bool check = top_layer->DeleteClusterMember(self);
  assert (check == true);

  int root1_index, root2_index;
  int * index_set1;
  int * index_set2;
  int set1_size, set2_size;

  // Partition the set of members
  int lower_size = top_layer->ag_list.GetSize() / 2;
  two_partition(top_layer->ag_list.GetSize(),cost,lower_size,index_set1,set1_size,root1_index,index_set2,set2_size,root2_index);
  free(cost);

#ifdef LOG_BSE_PARTITION
  printf ("[bse %d %d ] . . cluster-split : Partition 1 : ", id, n->id);
  for (int i = 0; i < set1_size; i++)
    printf ("%d ", ag_arr[index_set1[i]]->ag.agent_id);
  printf("\n");
  printf ("[bse %d %d ] . . cluster-split : Partition 2 : ", id, n->id);
  for (int i = 0; i < set2_size; i++)
    printf ("%d ", ag_arr[index_set2[i]]->ag.agent_id);
  printf("\n");
#endif // LOG_BSE_PARTITION
  
  LayerInfo * tmp_layer1 = new LayerInfo(top_layer->lid,NULL);
  LayerInfo * tmp_layer2 = new LayerInfo(top_layer->lid,NULL);
  for (int i = 0; i < set1_size; i++) {
    top_layer->DeleteClusterMember(ag_arr[index_set1[i]]);
    bool success = tmp_layer1->AddClusterMember(ag_arr[index_set1[i]]);
    assert (success == true);
  }
  tmp_layer1->root = ag_arr[root1_index];

  for (int i = 0; i < set2_size; i++) {
    top_layer->DeleteClusterMember(ag_arr[index_set2[i]]);
    bool success = tmp_layer2->AddClusterMember(ag_arr[index_set2[i]]);
    assert (success == true);
  }
  tmp_layer2->root = ag_arr[root2_index];

  free(index_set1);
  free(index_set2);

#ifdef LOG_BSE_PARTITION
  printf ("[bse %d %d ] At %8.4f cluster-split : root1 < [ %d %d ] > count %d     root2 < [ %d %d ] > count %d\n", id, n->id, Scheduler::Clock(), ag_arr[root1_index]->ag.agent_id, ag_arr[root1_index]->ag.node_id, set1_size, ag_arr[root2_index]->ag.agent_id, ag_arr[root2_index]->ag.node_id, set2_size);
#endif // LOG_BSE_PARTITION

  free(ag_arr);

  assert (top_layer->ag_list.GetSize() == 0);

  // Transfer control of the other cluster
  send_cluster_remove(tmp_layer1);
  send_cluster_remove(tmp_layer2);

  LayerAgentInfo * new_root1 = new LayerAgentInfo(tmp_layer1->root->ag.agent_id,tmp_layer1->root->ag.node_id);
  LayerAgentInfo * new_root2 = new LayerAgentInfo(tmp_layer2->root->ag.agent_id,tmp_layer2->root->ag.node_id);

  // Cleanup ...
  delete tmp_layer1;
  delete tmp_layer2;

  assert (top_layer->lid + 1 < MAX_LAYERS);
#ifdef LOG_BSE_CLUSTER_CHANGE
  printf ("[bse %d %d ] At %8.4f cluster-info : lid %d : deleted\n", id, n->id, Scheduler::Clock(), top_layer->lid);
#endif // LOG_BSE_CLUSTER_CHANGE
#ifdef LOG_BSE_JUNK
  printf ("[bse %d %d ] At %8.4f layer-increment : lid %d\n", id, n->id, Scheduler::Clock(), top_layer->lid+1);
#endif // LOG_BSE_JUNK

  // Setup the next higher layer
  top_layer->lid ++;
  top_layer->AddClusterRoot(self);
  top_layer->me_in_layer = true;

  top_layer->AddClusterMember(new_root1);
  top_layer->AddClusterMember(new_root2);

  return;
}

void bseAgent::send_cluster_remove (LayerInfo * l) {

#ifdef LOG_BSE_SEND
      printf ("[bse %d %d ] At %8.4f (b-scx) : self-cluster-remove : lid %d : new-root < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), top_layer->lid, l->root->ag.agent_id, l->root->ag.node_id);
#endif // LOG_BSE_SEND
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = l->ag_list.GetAt(pos);
    assert (self_check(la) == false);
    AppPacket *ap = new AppPacket (CLUSTER_REFRESH);
    put_cluster_remove_info_into_packet(l,ap,true,NULL,id,n->id);

#ifdef LOG_BSE_SEND
    printf ("[bse %d %d ] . . (b-scx) : < [ %d %d ] >\n", id, n->id, la->ag.agent_id, la->ag.node_id);
#endif // LOG_BSE_SEND

    send_pkt_wrapper(ap,la->ag.agent_id,la->ag.node_id);
  }
#ifdef LOG_BSE_SEND
  printf ("\n");
#endif // LOG_BSE_SEND
  return;
}

void bseAgent::log_cluster_change_info (void) {

  m_last_change = Scheduler::Clock();
  return;
}

void bseAgent::display_top_layer_info (void) {

  assert (top_layer != NULL);

  printf ("[bse %d %d ] At %8.4f cluster-info : lid %d : ldr %s %d : count %d : ", id, n->id, Scheduler::Clock(), top_layer->lid, (self_check(top_layer->root) == true) ? "self" : "other", top_layer->root->ag.agent_id, top_layer->ag_list.GetSize() );

#ifdef LOG_CLUSTER_INFO_DETAILED
  for (void * pos = top_layer->ag_list.GetHeadPosition();
       pos != NULL;
       top_layer->ag_list.GetNext(pos) ) {
    printf (" %d", (top_layer->ag_list.GetAt(pos))->ag.agent_id);
  }
#endif // LOG_CLUSTER_INFO_DETAILED

  printf ("\n");
  return;
}

bool bseAgent::self_check (LayerAgentInfo * la) {

  if ( (la->ag.agent_id == id) && (la->ag.node_id == n->id) )
    return true;
  else
    return false;
}
