/**************************************************
 *
  File: coop-agent.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  NICE implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#undef LOG_COOP_SEND
#undef LOG_COOP_RECV
#undef LOG_COOP_JUNK
#undef LOG_COOP_RECV_2

#undef LOG_COOP_JUNK_PING
#define LOG_COOP_JUNK_PING_SWITCH

#define LOG_COOP_START
#define LOG_COOP_STOP

#define LOG_COOP_CLUSTER_CHANGE
#undef  LOG_COOP_CLUSTER_PERIODIC
#undef LOG_CLUSTER_INFO_DETAILED

#define LOG_DATA_PACKET_INIT
#define LOG_DATA_PACKET_RECV
#define LOG_DATA_PACKET_SEND
#undef LOG_DATA_PACKET_SEND_COUNT

#define TESTING_JUNK

#include "nicenode.h"
#include <scheduler.h>
#include "app-packet.h"
#include "coop-agent.h"
#include "setpartition.h"
#include "common.h"

bool global_use_packet_cache = true;

coopAgent::coopAgent (void) : commonAgent () {

  jqt.ca = this;
  m_hlprt.ca = this;
  m_hlpit.ca = this;
}

coopAgent::~coopAgent (void) {

  if (started == true) {
    init_layer_arr();
    jqt.CancelTimer();
    cancel_higher_layer_ping_timers();
  }
}

void coopAgent::init (int Id, int Index, Node *N) {

  Agent::init(Id,Index,N);
  t = AGENT_APPLICATION_COOP;

  for (int i = 0; i < MAX_LAYERS; i++) {
    layers.arr[i] = NULL;
  }

  m_data_pkt_seq = 0;

  for (int i = 0; i < MAX_LAYERS; i++)
    m_last_change[i] = -1.0;

  return;
}

void coopAgent::start (void) {

  Agent::start();

  if (bse.agent_id == -1) {
    printf ("[Err] BSE is not known to agent %d\n", id);
    exit (-1);
  }
  state = INIT;

#ifdef LOG_COOP_START
  printf ("[coop %d %d ] At %8.4f start\n", id, n->id, Scheduler::Clock());
#endif // LOG_COOP_START

  init_layer_arr();
  init_join_query(0);

  init_packet_cache();
  m_use_packet_cache = global_use_packet_cache;
  m_ping_in_progress = false;
  return;
}

/* dst_join_lid < 0 => that attempt to join the same layer as before */
void coopAgent::init_join_query (int dst_join_lid) {

  if (dst_join_lid >= 0) {
    if (state == JOIN) {
      if (target_join_q_lid > dst_join_lid)
	target_join_q_lid = dst_join_lid;
      return;
    }
  }
  else {
    assert (state == JOIN);
  }
  assert (m_ping_in_progress == false);

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f init-join-query lid %d\n", id, n->id, Scheduler::Clock(), dst_join_lid);
#endif // LOG_COOP_JUNK
  state = JOIN;
  curr_join_q_lid = -1;
  if (dst_join_lid >= 0)
    target_join_q_lid = dst_join_lid;
  AppPacket *p = new AppPacket(JOIN_QUERY);
  init_join_query_packet(p,target_join_q_lid,Scheduler::Clock(),false);
  send_pkt_wrapper(p,bse.agent_id,bse.node_id);
  /* Set the timeout to resend the request if needed */
  jqt.SetTimer(CONST_JOIN_QUERY_TIMEOUT);
  return;
}

void coopAgent::init_layer_arr (void) {

  for (int i = 0; i < MAX_LAYERS; i++) {
    if (layers.arr[i] != NULL) {
      delete layers.arr[i];
      layers.arr[i] = NULL;
    }
  }
  return;
}

void coopAgent::stop (void) {

  assert (started == true);

#ifdef LOG_COOP_STOP
  printf ("[coop %d %d ] At %8.4f stop\n", id, n->id, Scheduler::Clock());
#endif // LOG_COOP_STOP

  jqt.CancelTimer();
  cancel_higher_layer_ping_timers();

  init_layer_arr();
  Agent::stop();
  return;
}

void coopAgent::specific_send_data_pkt (void) {

  assert (started == true);

#ifdef LOG_DATA_PACKET_INIT
  printf ("[coop %d %d ] At %8.4f source data-pkt : seq %d\n", id, n->id, Scheduler::Clock(), m_data_pkt_seq);
#endif // LOG_DATA_PACKET_INIT
  forward_data_packet(id, n->id, m_data_pkt_seq ++,-1,-1,true);
  return;
}

void coopAgent::specific_rx_pkt_handler  (Packet *p) {

  if (p->t != PACKET_APP) {

     printf ("[Err] COOP %d received unknown packet type\n", id);
     return;
  }

  AppPacket *ap = (AppPacket *)p;

  rx_pkt_wrapper(ap);

  switch (ap->st) {

  case JOIN_QUERY :
  case JOIN_FORWARD :

    handle_join_query_forward(ap);
    break;

  case JOIN_RESPONSE :

    handle_join_response(ap);
    break;

  case CLUSTER_REFRESH :

    handle_cluster_refresh(ap);
    break;

  case CLUSTER_MERGE :
    handle_cluster_merge(ap);
    break;

  case PING_QUERY :
    handle_ping_query(ap);
    break;

  case PING_RESPONSE :
    handle_ping_response(ap);
    break;

  case PACKET_DATA :
    handle_data_packet(ap);
    break;

  default :
    printf ("[Err] COOP %d received unknown packet subtype\n", id);
  }

  return;
}

void coopAgent::handle_join_query_forward (AppPacket * ap) {

  assert ( (ap->st == JOIN_QUERY) || (ap->st == JOIN_FORWARD) );

  int qlid;
  double src_time;
  bool attach;
  AgentInfo src_ag;
  AgentInfo original_dst;
  if (ap->st == JOIN_QUERY) {
    qlid = ap->u.joinq_p.q_lid;
    attach = ap->u.joinq_p.attach;
    src_ag.agent_id = ap->src_agent;
    src_ag.node_id = ap->src;
    original_dst.agent_id = id;
    original_dst.node_id = n->id;
    src_time = ap->u.joinq_p.send_time;
  }
  else {
    qlid = ap->u.joinforward_p.q_lid;
    attach = ap->u.joinforward_p.attach;
    src_ag.agent_id = ap->u.joinforward_p.src_ag.agent_id;
    src_ag.node_id = ap->u.joinforward_p.src_ag.node_id;
    original_dst.agent_id = ap->u.joinforward_p.original_dst.agent_id;
    original_dst.node_id = ap->u.joinforward_p.original_dst.node_id;
    src_time = ap->u.joinforward_p.send_time;
  }

  int required_qlid;
  if (attach == true)
    required_qlid = qlid;
  else
    required_qlid = qlid - 1;

  /* First check if this agent should send a response at all */
  if (valid_join_query_forward_packet(qlid,attach) == false) {
    AppPacket *resp_p = new AppPacket(JOIN_RESPONSE);
    resp_p->u.joinresp_p.layer_id = required_qlid;
    resp_p->u.joinresp_p.accept = false;
    resp_p->u.joinresp_p.mbr_count = 0;
    resp_p->u.joinresp_p.exp_src.agent_id = original_dst.agent_id;
    resp_p->u.joinresp_p.exp_src.node_id = original_dst.node_id;
    send_pkt_wrapper(resp_p,src_ag.agent_id, src_ag.node_id);
    return;
  }

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f (c-jqf) recvd-valid-join-qf : lid %d from < [ %d %d ] > exp-src < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), qlid,ap->src_agent,ap->src, src_ag.agent_id, src_ag.node_id);
#endif // LOG_COOP_JUNK

  LayerInfo * this_layer = layers.arr[required_qlid];

  if ( self_check(this_layer->root) == false) {
    /* If I am not the cluster root, then forward this to cluster root */
    AppPacket *fwd_p = new AppPacket(JOIN_FORWARD);
    put_info_into_join_forward_packet (src_ag,qlid,original_dst,src_time,attach,fwd_p);
    send_pkt_wrapper(fwd_p,this_layer->root->ag.agent_id, this_layer->root->ag.node_id);
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-jqf) : fwd to root < [ %d %d ] >\n", id, n->id, this_layer->root->ag.agent_id, this_layer->root->ag.node_id);
#endif // LOG_COOP_JUNK
    return;
  }

  if (attach == false) {

    void * pos = this_layer->ag_list.Locate(src_ag.agent_id);
    if (pos != NULL) {
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-jqf) : dup delete\n", id, n->id);
#endif // LOG_COOP_JUNK
      LayerAgentInfo * this_la = this_layer->ag_list.GetAt(pos);
      assert (this_la != NULL);
      this_layer->ag_list.RemoveAt(pos);
      delete this_la;
      log_cluster_change_info(this_layer->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
      display_cluster_info(this_layer,this_layer->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
    }

    /* Create response packet */
    AppPacket *resp_p = new AppPacket(JOIN_RESPONSE);
    put_layer_cluster_info_into_packet(this_layer,resp_p);
    resp_p->u.joinresp_p.accept = true;
    resp_p->u.joinresp_p.exp_src.agent_id = original_dst.agent_id;
    resp_p->u.joinresp_p.exp_src.node_id = original_dst.node_id;
    send_pkt_wrapper(resp_p,src_ag.agent_id, src_ag.node_id);
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-jqf) : sent response\n", id, n->id);
#endif // LOG_COOP_JUNK
  }

  else { /* Attach appropriately to the cluster */
      LayerAgentInfo *la_ag = new LayerAgentInfo(src_ag.agent_id,src_ag.node_id);
      if (this_layer->AddClusterMember(la_ag) == false) {
	delete la_ag;
	la_ag = this_layer->FindClusterMember(src_ag.agent_id);
	la_ag->refresh = true;
      }
      // if (ap->st == JOIN_QUERY) { // and not JOIN_FORWARD
      la_ag->dist = Scheduler::Clock() - src_time;
      // }

#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-jqf) : *attach*\n", id, n->id);
#endif // LOG_COOP_JUNK

      // Immediate update of new member joining the cluster
      this_layer->cr_msgt.CancelTimer();
      send_all_cluster_refresh_packet(this_layer,false,true,NULL);
      this_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  }

  return;
}

void coopAgent::split_cluster_using_two_partition (LayerInfo * l) {


  LayerAgentInfo ** ag_arr = l->CreateLayerAgentInfoArray();
  assert (ag_arr != NULL);
  double * cost = create_cost_matrix(id,n->id,ag_arr,l->ag_list.GetSize());
  int root1_index, root2_index;
  int * index_set1 = NULL;
  int * index_set2 = NULL;
  int set1_size, set2_size;

  // Partition the set of members
  int lower_size = l->ag_list.GetSize() / 2;
  two_partition(l->ag_list.GetSize(),cost,lower_size,index_set1,set1_size,root1_index,index_set2,set2_size,root2_index);
  assert (index_set1 != NULL);
  assert (index_set2 != NULL);
  free(cost);

  // After the split, cluster1 is my cluster, cluster2 is the other cluster
  LayerAgentInfo ** cluster1 = (LayerAgentInfo **) safe_malloc (sizeof (LayerAgentInfo *) * set1_size);
  bool need_to_swap = true;
  for (int i = 0; i < set1_size; i++) {
    cluster1[i] = ag_arr[index_set1[i]];
    if (self_check(cluster1[i]) == true)
      need_to_swap = false;
  }
  LayerAgentInfo ** cluster2 = (LayerAgentInfo **) safe_malloc (sizeof (LayerAgentInfo *) * set2_size);
  for (int i = 0; i < set2_size; i++)
    cluster2[i] = ag_arr[index_set2[i]];

  LayerAgentInfo * c1_root = ag_arr[root1_index];
  LayerAgentInfo * c2_root = ag_arr[root2_index];

  if (need_to_swap == true) {
    LayerAgentInfo ** tmp_cl = cluster1;
    cluster1 = cluster2;
    cluster2 = tmp_cl;
    int tmp_cl_size = set1_size;
    set1_size = set2_size;
    set2_size = tmp_cl_size;
    LayerAgentInfo * tmp_root = c1_root;
    c1_root = c2_root;
    c2_root = tmp_root;
  }

  free(ag_arr);
  free(index_set1);
  free(index_set2);

  // Purge my own cluster of members that are in the other cluster
  // Create a temporary layer to send out the root transfer message
  LayerInfo * tmp_layer = new LayerInfo (l->lid,this);
  for (int i = 0; i < set2_size; i++) {
    tmp_layer->AddClusterMember(cluster2[i]);
    l->DeleteClusterMember(cluster2[i]);
  }
  tmp_layer->root = c2_root;

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f split-cluster-self lid %d\n", id, n->id, Scheduler::Clock(), l->lid);
#endif // LOG_COOP_JUNK

  // Send refresh or root transfer for my cluster
  if (self_check(c1_root) == false) {
    do_internal_root_transfer(l,c1_root);
  }
  else {
    l->cr_msgt.CancelTimer();
    send_all_cluster_refresh_packet(l,false,true,NULL);
    l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  }

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f split-cluster-other lid %d\n", id, n->id, Scheduler::Clock(), l->lid);
#endif // LOG_COOP_JUNK
  // Transfer control of the other cluster
  send_cluster_remove(tmp_layer,true);

  // Cleanup ...
  delete tmp_layer;
  free(cluster1);
  free(cluster2);

  return;
}


/* This will split the cluster associated with lid. The new cluster
 * that contains this member will stay as is, the other cluster
 * will be spawned off.
 */
void coopAgent::split_cluster_assuming_2k (LayerInfo * l) {

  // Currently, I am implementing a naive scheme. The leader
  // chooses its nearest neighbor to stay with it in its cluster.
  // Other more intelligent schemes will be implemented later.

  assert (l->ag_list.GetSize() == UPPER_2K);
  LinkedList<LayerAgentInfo *, double> my_cluster_list;
  /* In my_cluster, the agents are sorted in decreasing order of distance */
  LinkedList<LayerAgentInfo *, double> other_cluster_list;

  /* First, split the agents into my cluster and other cluster */
  /* All agents that I know distances to, are my cluster to start with */
  LayerAgentInfo *self = NULL;
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo *la_ag = l->ag_list.GetAt(pos);
    if ( self_check(la_ag) == true) {
      self = la_ag;
      continue;
    }
    if (la_ag->dist >= 0.0) {
      /* -1.0 is for sorting it in decreasing order of distance */
      my_cluster_list.Add(la_ag,-1.0 * la_ag->dist);
    }
    else { /* Sorted by increasing distance order */
      other_cluster_list.Add(la_ag,la_ag->dist);
    }
  }

  /* Whichever set has the higher number of agents, need to transfer
   * some agents to the other set */
  void *pos2 = other_cluster_list.GetHeadPosition();
  while ( my_cluster_list.GetSize() < (LOWER_K - 1) ) {
    
    LayerAgentInfo *a = other_cluster_list.GetAt(pos2);
    my_cluster_list.Add(a,-1.0 * a->dist);
    void * old_pos2 = pos2;
    other_cluster_list.GetNext(pos2);
    other_cluster_list.RemoveAt(old_pos2);
  }
  
  void *pos3 = my_cluster_list.GetHeadPosition();
  while (other_cluster_list.GetSize() < LOWER_K) {
    
    LayerAgentInfo *a = my_cluster_list.GetAt(pos2);
    other_cluster_list.Add(a,a->dist);
    void * old_pos3 = pos3;
    my_cluster_list.GetNext(pos3);
    my_cluster_list.RemoveAt(old_pos3);
  }

  // Re-create my own cluster tree
  l->RemoveAgents();
  l->AddClusterRoot(self);
  l->me_in_layer = true;
  for (void *pos4 = my_cluster_list.GetHeadPosition();
       pos4 != NULL;
       my_cluster_list.GetNext(pos4) ) {

    l->AddClusterMember(my_cluster_list.GetAt(pos4));
  }
  // Create the other cluster
  // SimpleTree<LayerAgentInfo *> other_cluster_tree;
  // void * pos5 = other_cluster_list.GetHeadPosition();
  // assert (pos5 != NULL);
  // LayerAgentInfo * oc_root_la = other_cluster.GetAt(pos5);
  // void * oc_root_pos = other_cluster_tree.SetRoot(oc_root_la);
  // for (pos5 = other_cluster_list.GetNext(pos5);
  //      pos5 != NULL;
  //      other_cluster_list.GetNext(pos5) ) {

  //   other_cluster_tree.AddChild(other_cluster_list.GetAt(pos5),oc_root_pos);
  // }

  /* HAVE TO SEND MESSAGE TO THE OTHER CLUSTER MEMBERS */

  my_cluster_list.RemoveAll();
  other_cluster_list.RemoveAll();

  return;
}

void coopAgent::handle_join_response (AppPacket * ap) {

  /* Need to validate if this is a desired response packet */
  if (valid_join_response_packet(ap) == false)
    return;

  int this_lid = ap->u.joinresp_p.layer_id;

#ifdef LOG_COOP_RECV
  printf ("\n[coop %d %d ] At %8.4f (c-jr) : join-resp : lid %d from < [ %d %d ] > exp-src < [ %d %d ] >\n", id, n->id, Scheduler::Clock(),this_lid,ap->src_agent,ap->src, ap->u.joinresp_p.exp_src.agent_id, ap->u.joinresp_p.exp_src.node_id);
#endif // LOG_COOP_RECV

  if ( (ap->src_agent == bse.agent_id) && (ap->src == bse.node_id) ) {
 
    assert (ap->u.joinresp_p.accept == true);

    // BSE-JF : Have to delete all other members in curr_join_q_lid etc.
    // or have to wait till all resps are recvd and decide whether to
    // accept BSE or the other members

    flush_layer(this_lid);
    if (this_lid+1 < MAX_LAYERS) {
      flush_layer(this_lid+1);
    }

    layers.arr[this_lid] = new LayerInfo(this_lid,this);
    LayerInfo * this_layer = layers.arr[this_lid];
    jqt.CancelTimer();

    /* This is the top layer returned by BSE */
    put_packet_cluster_info_into_layer(ap,this_layer);
    if (this_lid == target_join_q_lid) { // This is the desired layer
      void * pos = this_layer->ag_list.Locate(id);
      assert (pos != NULL);
      LayerAgentInfo * self = this_layer->ag_list.GetAt(pos);
      self->dist = 0.0;

      LayerAgentInfo * bse_la = this_layer->FindClusterMember(bse.agent_id);
      assert (bse_la != NULL);
      bse_la->dist = Scheduler::Clock() - ap->send_time;

      this_layer->me_in_layer = true;
      this_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
      this_layer->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);

      state = ATTACHED;

      set_higher_layer_ping_timer();

#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-jr) : attach-to-bse\n\n", id, n->id);
#endif // LOG_COOP_JUNK
      log_cluster_change_info(this_layer->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
      display_cluster_info(this_layer,this_layer->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
    }
    else {
      assert (this_lid > target_join_q_lid);

      void * pos = this_layer->ag_list.Locate(id);
      if (pos != NULL) { // I am considered to be in this cluster !
	assert ( self_check(this_layer->root) == false); // I am not root	
#ifdef LOG_COOP_JUNK
        printf ("[coop %d %d ] . . (c-jr) : part of returned cluster! delete\n", id, n->id);
#endif // LOG_COOP_JUNK
	delete_self_from_layer(this_layer);
      }
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-jr) : continue-join-query\n\n", id, n->id);
#endif // LOG_COOP_JUNK
      continue_join_query(this_lid);
    }
  }
  else { /* This is a message from the higher layer cluster
	  * We need to just save the information till the timeout
	  */
    LayerInfo * higher_layer = layers.arr[this_lid + 1];
    assert (higher_layer != NULL);
    void * mem_pos = higher_layer->ag_list.Locate(ap->u.joinresp_p.exp_src.agent_id);
    assert (mem_pos != NULL);
    LayerAgentInfo *hl_member = higher_layer->ag_list.GetAt(mem_pos);

    if (ap->u.joinresp_p.accept == false) {
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-jr) : join-reject\n\n", id, n->id);
#endif // LOG_COOP_JUNK
      higher_layer->DeleteClusterMember(hl_member);
      delete hl_member;
    }
    else {
      /* If the response is coming from a different agent (I thought some
       * one else was the cluster leader, then we need to update that here */
      if ( (ap->u.joinresp_p.exp_src.agent_id != ap->src_agent) || (ap->u.joinresp_p.exp_src.node_id != ap->src) ) {
	hl_member->ag.agent_id = ap->src_agent;
	hl_member->ag.node_id = ap->src;
      }

      assert (ap->u.joinresp_p.mbr_count > 0);
      assert ( (hl_member->ag.agent_id == ap->u.joinresp_p.agent_arr[0].ag.agent_id) && (hl_member->ag.node_id == ap->u.joinresp_p.agent_arr[0].ag.node_id) );

      if (hl_member->lower_layer != NULL)
	delete hl_member->lower_layer;
      else
	m_join_resp_received ++;
      
      hl_member->lower_layer = new LayerInfo (this_lid,this);
      put_packet_cluster_info_into_layer (ap,hl_member->lower_layer);

      LayerAgentInfo * self = hl_member->lower_layer->FindClusterMember(id);
      if (self != NULL)
	self->dist = 0.0;

      /* Using the time metric. Can use any other */
      hl_member->dist = Scheduler::Clock() - ap->send_time;
    }

    int match_count = higher_layer->ag_list.GetSize();
    if (higher_layer->ag_list.Locate(bse.agent_id) != NULL)
      match_count --;
    if (match_count == m_join_resp_received) {
      
      jqt.CancelTimer();
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-jr) : all join resps found\n\n", id, n->id);
#endif // LOG_COOP_JUNK
      process_received_join_responses();
    }
  }
  return;
}

void coopAgent::handle_cluster_refresh (AppPacket *ap) {

  assert (ap->st == CLUSTER_REFRESH);
  if (valid_cluster_refresh_packet(ap) == false)
    return;
  LayerInfo * l = layers.arr[ap->u.clusterrefresh_p.layer_id];
  void * this_agent_pos = l->ag_list.Locate(ap->src_agent); 
  assert (this_agent_pos != NULL);
  LayerAgentInfo *la_ag = l->ag_list.GetAt(this_agent_pos);;

  bool test_for_all_pings_received = false;

  bool change = false;
  bool match_to_root_done = false;

#ifdef TESTING_JUNK
  void * pos = l->ag_list.Locate(id);
  if (pos != NULL) {
    LayerAgentInfo * self = l->ag_list.GetAt(pos);
    assert (self->dist >= 0.0);
  }
#endif // TESTING_JUNK

#ifdef LOG_COOP_RECV
  printf ("\n[coop %d %d ] At %8.4f (c-rcr) : recvd-valid-refresh : lid %d from < [ %d %d ] > %s %s %s\n", id, n->id, Scheduler::Clock(), l->lid,ap->src_agent,ap->src, (ap->u.clusterrefresh_p.is_root == true) ? "root" : ".", (ap->u.clusterrefresh_p.root_xfer == true) ? "xfer" : ".", (ap->u.clusterrefresh_p.cluster_remove == true) ? "remove" : ".");
#endif // LOG_COOP_RECV

#ifdef LOG_COOP_RECV_2
  printf ("[coop %d %d ] . . (c-rcr) : dist-info in refresh :", id, n->id);
  for (int i = 0; i < ap->u.clusterrefresh_p.mbr_count; i++)
    printf ("( %d %5.3f )", ap->u.clusterrefresh_p.agent_arr[i].ag.agent_id,ap->u.clusterrefresh_p.agent_arr[i].dist);
  printf ("\n");
#endif // LOG_COOP_RECV_2

  /* Now this is a valid packet */
  la_ag->refresh_agent(ap);

  assert (l->root != NULL);
  if (self_check(l->root) == true) { // I am the root
    if (ap->u.clusterrefresh_p.is_root == true) {

    // Pkt src,s, thinks that it is the root

      // First check if it is doing a root transfer to me
      if ( (ap->u.clusterrefresh_p.root_xfer == true) &&
	   ( (ap->u.clusterrefresh_p.agent_arr[0].ag.agent_id == id) && 
	     (ap->u.clusterrefresh_p.agent_arr[0].ag.node_id == n->id) ) ) {
        // In which case gladly add all unknown members
        bool members_added = false;
	if (l->AddExtraMembersFromAltRootPacket (ap) > 0) {
	  change = true;
          members_added = true;
        }
	// Do cluster remove
	if (ap->u.clusterrefresh_p.cluster_remove == true) { // Delete
#ifdef LOG_COOP_JUNK
          printf ("[coop %d %d ] . . (c-rcr) : src self-remove\n", id, n->id);
#endif // LOG_COOP_JUNK
	  bool check = l->DeleteClusterMember(la_ag);
	  assert (check == true);
	  delete la_ag;
	  change = true;
	}
#ifdef LOG_COOP_JUNK
        printf ("[coop %d %d ] . . (c-rcr) : src and me roots, root xfer to me\n\n", id, n->id);
#endif // LOG_COOP_JUNK
	if (change == true) {
          log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
	  display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
        }

	// Immediate update of new member additions
        if (members_added == true) {
          l->cr_msgt.CancelTimer();
          send_all_cluster_refresh_packet(l,false,true,NULL);
          l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
	}
	return;
      }

      bool stay_as_root = false;

      // s_max: source dist to its farthest member
      double s_max = get_farthest_dist_of_member_to_agent(l,la_ag,false,false);
      LayerAgentInfo * my_farthest_ag = find_farthest_agent_to_me_in_layer(l);
      double my_max = my_farthest_ag->dist;

      long s_max_long = USECS(s_max);
      long my_max_long = USECS(my_max);

      int s_know_count = count_dist_known_members_for_agent(la_ag,false);
      int me_know_count = count_dist_known_members_for_me(l);
      if (are_agent_members_contained_in_mine(l,la_ag) == true) {
	// All members in la_ag are also known to me
	if ( (s_know_count < me_know_count) || // I am the superset
	     (my_max_long < s_max_long) || // We are equal in size, but I am better
	     ( (my_max_long == s_max_long) && (id < la_ag->ag.agent_id) ) ) { // -do-
	  stay_as_root = true;
	} // else {stay_as_root = false;}
      }
      else if (are_my_members_contained_in_agent(l,la_ag) == false) { 
	// no one is a superset of the other
	if (id < la_ag->ag.agent_id) {
	  stay_as_root = true;
	} // else {stay_as_root = false;}

      } // else my_members_contained_in_agent() == true: 
        // i.e. s is a superset, so stay_as_root = false

      LayerAgentInfo * other_root;
      bool member_added = false;
      if (ap->u.clusterrefresh_p.root_xfer == true) {
	other_root = l->FindClusterMember(ap->u.clusterrefresh_p.agent_arr[0].ag.agent_id);
	if (other_root == NULL) {
	  LayerAgentInfo *new_guy = new LayerAgentInfo(ap->u.clusterrefresh_p.agent_arr[0].ag.agent_id,ap->u.clusterrefresh_p.agent_arr[0].ag.node_id);
	  l->AddClusterMember(new_guy);
	  other_root = new_guy;
	  change = true;
	  member_added = true;
	}
#ifdef LOG_COOP_JUNK
        printf ("[coop %d %d ] . . (c-rcr) : src and me roots, src root xfer to < [ %d %d ] >\n", id, n->id, other_root->ag.agent_id, other_root->ag.node_id);
#endif // LOG_COOP_JUNK
      }
      else {
	assert (ap->u.clusterrefresh_p.cluster_remove == false);
	other_root = la_ag;
      }

      if (stay_as_root == true) {
	// Add all the extra members and send root challenge
	if (l->AddExtraMembersFromAltRootPacket(ap) > 0) {
	  change = true;
          member_added = true;
        }
	if (ap->u.clusterrefresh_p.cluster_remove == true) { // Delete
#ifdef LOG_COOP_JUNK
          printf ("[coop %d %d ] . . (c-rcr) : src self-remove\n", id, n->id);
#endif // LOG_COOP_JUNK
	  bool check = l->DeleteClusterMember(la_ag);
	  assert (check == true);
	  delete la_ag;
	  change = true;
	}
#ifdef LOG_COOP_JUNK
        printf ("[coop %d %d ] . . (c-rcr) : me stay as root (root-challenge)\n\n", id, n->id);
#endif // LOG_COOP_JUNK
	send_root_challenge(l,other_root);
	if (change == true) {
          log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
	  display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
        }

	// Immediate inform other members of new member
	if (member_added == true) {
          l->cr_msgt.CancelTimer();
	  // Do not send this to the other root
          send_all_cluster_refresh_packet(l,false,true,other_root);
          l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
        }

	return;
      }
      else {
	change = true;
	if (ap->u.clusterrefresh_p.cluster_remove == true) { // Delete
#ifdef LOG_COOP_JUNK
          printf ("[coop %d %d ] . . (c-rcr) : src self-remove\n", id, n->id);
#endif // LOG_COOP_JUNK
	  bool check = l->DeleteClusterMember(la_ag);
	  assert (check == true);
	  delete la_ag;
	}
#ifdef LOG_COOP_JUNK
        printf ("[coop %d %d ] . . (c-rcr) : src stay as root (root-xfer to src)\n\n", id, n->id);
#endif // LOG_COOP_JUNK
	do_internal_root_transfer(l,other_root);

        log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
	display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
	
	return;
      }
    }
    // else(s does not think it is the root): nothing to do
  }
  else { // I am not the root
    AgentInfo old_root(l->root->ag.agent_id,l->root->ag.node_id);

    if ( (l->root->ag.agent_id == ap->src_agent) && 
	 (l->root->ag.node_id == ap->src) ) { // I think s is the root
      if (match_cluster_to_root_info(l) == true) {
	change = true;
	match_to_root_done = true;
      }
      if ( (l->lid + 1) < MAX_LAYERS) {
        if (layers.arr[l->lid+1] == NULL)
          layers.arr[l->lid+1] = new LayerInfo (l->lid+1, this);
	put_cluster_refresh_hl_info_into_layer_agent(ap, layers.arr[l->lid+1],id,n->id);
	if (m_ping_in_progress == true) {
	//  m_ping_resps_received = get_ping_response_count_in_layer(layers.arr[l->lid+1]);
	  test_for_all_pings_received = true;
        }
      }

      if ( ( (ap->u.clusterrefresh_p.is_root == true) && // s thinks it is root
	     (ap->u.clusterrefresh_p.root_xfer == true) ) ||
	   (ap->u.clusterrefresh_p.is_root == false) ) { // s doesnt think it is root
	void *pos = l->ag_list.Locate(ap->u.clusterrefresh_p.agent_arr[0].ag.agent_id);
	if (pos == NULL)
	  assert (pos != NULL);
	LayerAgentInfo *new_root = l->ag_list.GetAt(pos);

	assert ( (old_root.agent_id != new_root->ag.agent_id) || (old_root.node_id != new_root->ag.node_id) ); // THE IF STMT (NEXT LINE) IS NOT NEEDED
	if ( (old_root.agent_id != new_root->ag.agent_id) || (old_root.node_id != new_root->ag.node_id) ) {
#ifdef LOG_COOP_JUNK
          printf ("[coop %d %d ] . . (c-rcr) : have a new cluster-root < [ %d %d ] >\n", id, n->id, new_root->ag.agent_id, new_root->ag.node_id);
#endif // LOG_COOP_JUNK
	  l->root = new_root;
	  change = true;

	  cancel_higher_layer_ping_timers();
	}
	// Join next higher layer, if it was a root transfer to me
	if (self_check(l->root) == true) {
#ifdef LOG_COOP_JUNK
          printf ("[coop %d %d ] . . (c-rcr) : me not root, root xfer to me\n", id, n->id);
#endif // LOG_COOP_JUNK
	  join_to_next_higher_layer(l->lid);
	  l->delay_root_xfer = true;
        }
	else {
	  // The higher layer ping timer is set only when the next higher
	  // layer info is known. If I am joining as the new leader, then
	  // this info is not known.
	  set_higher_layer_ping_timer();
	}
      }
    } // else I dont think s is the root: do nothing
  }

  if (match_to_root_done == false) { // Only in this case, is this check
  // required to see if we need to delete and free this member. In all
  // other cases, we have already done the delete and returned from the
  // function.
  // So if match_to_root not done, and control is here, then do this
  // check and delete if appropriate.
    if (ap->u.clusterrefresh_p.cluster_remove == true) { // Delete
      bool check = l->DeleteClusterMember(la_ag);

      assert (check == true); // This is true after the match_to_root stuff.

      if (check == true) {
#ifdef LOG_COOP_JUNK
        printf ("[coop %d %d ] . . (c-rcr) : src self-remove\n", id, n->id);
#endif // LOG_COOP_JUNK
        delete la_ag;
        change = true;
      }
    }
  }

  if ( (m_ping_in_progress == true) && (test_for_all_pings_received == true) ) {
    m_ping_resps_received = get_ping_response_count_in_layer(layers.arr[l->lid+1]);
    if ( (l->lid + 1) != m_ping_lid )
      assert ( (l->lid+1) == m_ping_lid);
    if (check_all_ping_responses_received() == true) {
      int lower_lid = m_ping_lid - 1;
      //printf ("[%d %d] At %8.4f ping before-process handle-cluster-refresh\n", id, n->id, Scheduler::Clock());
      if (process_ping_responses () == true) {

	//printf ("[%d %d] At %8.4f ping handle-cluster-refresh\n", id, n->id, Scheduler::Clock());
      	change = true;
        l = layers.arr[lower_lid];
      }
    }
  }

  if (change == true) {
    assert (l != NULL);
    log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
    display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
  }

  return;
}

void coopAgent::handle_cluster_merge (AppPacket * ap) {

  if (valid_cluster_merge_packet(ap) == false)
    return;
#ifdef LOG_COOP_RECV
  printf ("[coop %d %d ] At %8.4f (c-merge) : recvd-valid-merge : lid %d from < [ %d %d ]>\n",  id, n->id,Scheduler::Clock(),ap->u.clustermerge_p.layer_id,ap->src_agent,ap->src);
  for (int i = 0; i < ap->u.clustermerge_p.mbr_count; i++)
    printf ("[coop %d %d ] . . (c-merge) < [ %d %d ] >\n", id, n->id, ap->u.clustermerge_p.agent_arr[i].ag.agent_id,  ap->u.clustermerge_p.agent_arr[i].ag.node_id);
  printf ("\n");
#endif // LOG_COOP_RECV

  LayerInfo *l = layers.arr[ap->u.clustermerge_p.layer_id];
  assert (l != NULL);
  int count = l->AddExtraMembersFromAltRootPacket(ap);
  if (count > 0) {
    log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
    display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
  }
  l->cr_msgt.CancelTimer();
  send_all_cluster_refresh_packet(l,false,true,NULL);
  l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
  return;
}

void coopAgent::handle_ping_query (AppPacket *ap) {
  
  assert (ap->st == PING_QUERY);
  int this_lid = ap->u.pingq_p.lid;
  bool accept = false;

#ifdef LOG_COOP_JUNK_PING
  printf ("[coop %d %d ] At %8.4f recv-pingq : lid %d : from < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), ap->u.pingq_p.lid, ap->src_agent, ap->src);
#endif // LOG_COOP_JUNK_PING

  if (layers.arr[this_lid] != NULL) {
    if (layers.arr[this_lid]->me_in_layer == true) {

      /* This is new stuff: reject a ping request if accepting this
       * member increases my cluster radius */
      LayerInfo * lower_layer = layers.arr[this_lid - 1];
      assert (lower_layer != NULL);
      assert (lower_layer->root != NULL);
      assert (self_check(lower_layer->root) == true);
      LayerAgentInfo * my_farthest = find_farthest_agent_to_me_in_layer(lower_layer);
      assert (my_farthest != NULL);
      long my_farthest_dist_long = USECS(my_farthest->dist);
      long this_dist_long = USECS(Scheduler::Clock() - ap->send_time);
      if (my_farthest_dist_long > this_dist_long)
        accept = true;
      else {
#ifdef LOG_COOP_JUNK_PING
  	printf ("[coop %d %d ] . . recv-pingq : reject (dist-violation)\n", id, n->id);
#endif // LOG_COOP_JUNK_PING
      }
    }
  }
  AppPacket *resp_p = new AppPacket (PING_RESPONSE);
  resp_p->u.pingresp_p.accept = accept;
  resp_p->u.pingresp_p.lid = this_lid;
  resp_p->u.pingresp_p.src_time = ap->u.pingq_p.src_time;

  if (accept == true) {
    assert (this_lid > 0);
    LayerInfo * lower_layer = layers.arr[this_lid - 1];
    //assert (lower_layer != NULL);
    //assert (lower_layer->root != NULL);
    //assert (self_check(lower_layer->root) == true);

    void * pos = lower_layer->ag_list.Locate(ap->src_agent);
    if (pos != NULL) {
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] At %8.4f pinger < [ %d %d ] > dup\n", id, n->id, Scheduler::Clock(),ap->src_agent,ap->src);
#endif // LOG_COOP_JUNK
      LayerAgentInfo * this_la = lower_layer->ag_list.GetAt(pos);
      lower_layer->ag_list.RemoveAt(pos);
      delete this_la;
      log_cluster_change_info(lower_layer->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
    display_cluster_info(lower_layer,lower_layer->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
    }

    put_layer_cluster_info_into_packet(lower_layer,resp_p);
  }

  send_pkt_wrapper(resp_p,ap->src_agent,ap->src);
  return;
}

void coopAgent::handle_ping_response (AppPacket * ap) {

  assert (ap->st == PING_RESPONSE);

  if ( (m_ping_in_progress == false) || (ap->u.pingresp_p.lid != m_ping_lid) )
    return;
  LayerInfo * l = layers.arr[m_ping_lid];
  assert (l != NULL);
  void * pos = l->ag_list.Locate(ap->src_agent);
  if (pos == NULL)
    return;
  LayerAgentInfo * la = l->ag_list.GetAt(pos);
  la->dist = (Scheduler::Clock() - ap->u.pingresp_p.src_time) / 2.0;


  if (la->lower_layer != NULL) {
    delete la->lower_layer;
    la->lower_layer = NULL;
  }
#ifdef LOG_COOP_JUNK_PING
  printf ("[coop %d %d ] At %8.4f recv-pingresp : lid %d : from < [ %d %d ] > dist %f %s\n", id, n->id, Scheduler::Clock(), m_ping_lid, ap->src_agent, ap->src, la->dist, (ap->u.pingresp_p.accept == true) ? "" : "*reject*" );
#endif // LOG_COOP_JUNK_PING
  if (ap->u.pingresp_p.accept == true) {
    la->lower_layer = new LayerInfo(m_ping_lid-1,this);
    put_packet_cluster_info_into_layer(ap,la->lower_layer);
    m_ping_resps_received ++;
  }
  else {
    bool success = l->DeleteClusterMember(la);
    delete la;
    assert (success == true);
  }

  if (check_all_ping_responses_received() == true) {
    int lower_lid = m_ping_lid - 1;
    if (process_ping_responses () == true) {
      //printf ("[%d %d] At %8.4f ping recv-ping-resp\n", id, n->id, Scheduler::Clock());
      LayerInfo * new_layer = layers.arr[lower_lid];
      assert (new_layer != NULL);
      log_cluster_change_info(new_layer->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
      display_cluster_info(new_layer,new_layer->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
    }
  }

  return;
}

bool coopAgent::check_all_ping_responses_received (void) {

  assert (m_ping_in_progress == true);

  LayerInfo *l = layers.arr[m_ping_lid];
  assert (l != NULL);
  // If this higher layer has BSE, then cannot count it, since BSE is not sent
  // a ping.
  int ignore_bse;
  if (l->ag_list.Locate(bse.agent_id) != NULL)
    ignore_bse = 1;
  else
    ignore_bse = 0;
  // The -1 is due to not including the root of the lower layer in this
  if ( m_ping_resps_received == ( l->ag_list.GetSize() - 1 - ignore_bse) )
    return true;
  return false;
}

/* Returns true if there is a change */
bool coopAgent::process_ping_responses (void) {

  bool ret_val = false;

  assert (m_ping_lid > 0);
  LayerInfo * hl = layers.arr[m_ping_lid];
  assert (hl != NULL);

  LayerInfo * my_highest_layer = layers.arr[m_ping_lid - 1];
  assert (my_highest_layer != NULL);
  assert (my_highest_layer->root != NULL);
  if (my_highest_layer->root->dist < 0.0)
    assert (my_highest_layer->root->dist >= 0.0);
  assert (my_highest_layer->me_in_layer == true);

  LayerAgentInfo * min_la = find_closest_agent_in_layer(hl,&(my_highest_layer->root->ag),true);

  if (min_la != NULL) {
    long my_highest_layer_root_dist_long = USECS(my_highest_layer->root->dist);
    long min_la_dist_long = USECS(min_la->dist);
    if (my_highest_layer_root_dist_long > min_la_dist_long) {

      ret_val = true;
#ifdef LOG_COOP_JUNK_PING_SWITCH
      printf ("[coop %d %d ] At %8.4f switch-root : lid %d : from < [ %d %d ] > to < [ %d %d ] > dist %8.4f \n", id, n->id, Scheduler::Clock(), m_ping_lid - 1, my_highest_layer->root->ag.agent_id,my_highest_layer->root->ag.node_id, min_la->ag.agent_id, min_la->ag.node_id, min_la->dist);
#endif // LOG_COOP_JUNK_PING_SWITCH

      // Send the cluster remove in old cluster
      // Delete self
      void * pos = my_highest_layer->ag_list.Locate(id);
      assert (pos != NULL);
      LayerAgentInfo * self_ag = my_highest_layer->ag_list.GetAt(pos);
      my_highest_layer->ag_list.RemoveAt(pos);
      my_highest_layer->me_in_layer = false;
  
      if (my_highest_layer->ag_list.GetSize() > 0)
	send_cluster_remove(my_highest_layer,false);

      delete my_highest_layer;

      // Send the join to new cluster
      layers.arr[m_ping_lid-1] = min_la->lower_layer;
      min_la->lower_layer = NULL;

      LayerInfo * new_layer = layers.arr[m_ping_lid-1];
      bool check = new_layer->AddClusterMember(self_ag);
      assert (check == true);
      new_layer->me_in_layer = true;
    
      assert (new_layer->root != NULL);

      assert ( (new_layer->root->ag.agent_id == min_la->ag.agent_id) && (new_layer->root->ag.node_id == min_la->ag.node_id) );
      assert (min_la->dist >= 0.0);
      new_layer->root->dist = min_la->dist;

      AppPacket *pkt = new AppPacket(JOIN_QUERY);
      init_join_query_packet(pkt,m_ping_lid-1,Scheduler::Clock(),true);
      send_pkt_wrapper(pkt,new_layer->root->ag.agent_id, new_layer->root->ag.node_id);
      new_layer->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
      new_layer->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);
      assert (state == ATTACHED);

    }
  }

  cancel_higher_layer_ping_timers();
  set_higher_layer_ping_timer();
  return ret_val;
}

void coopAgent::handle_data_packet (AppPacket *ap) {

  assert (ap->st == PACKET_DATA);
#ifdef LOG_DATA_PACKET_RECV
  printf ("[coop %d %d ] At %8.4f recv data-pkt : < [ %d %d ] > seq %d : from < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), ap->u.data_p.original_src.agent_id, ap->u.data_p.original_src.node_id, ap->u.data_p.seq_no, ap->src_agent, ap->src);
#endif // LOG_DATA_PACKET_RECV

  if (m_use_packet_cache == true) {

    if (add_to_packet_cache(ap->u.data_p.original_src.agent_id,ap->u.data_p.original_src.node_id,ap->u.data_p.seq_no) == false)
      return;
  }

  int this_lid = ap->u.data_p.lid;
  LayerInfo *l = layers.arr[this_lid];
  /* The 3 if conditions below enforce strict checking of packet source */
  if (l != NULL) {
    if (l->me_in_layer == true) {
      // Somewhat looser checking of data packet forwarding
      // if (l->ag_list.Locate(ap->src_agent) != NULL) {

        forward_data_packet(ap->u.data_p.original_src.agent_id, ap->u.data_p.original_src.node_id, ap->u.data_p.seq_no, ap->src_agent, ap->src,false);
      // }
    }
  }

  return;
}

void coopAgent::forward_data_packet (int src_aid, int src_nid, int src_seqno, int pred_aid, int pred_nid, bool is_source) {

  int count_pkts = 0;

  for (int i = 0; i < MAX_LAYERS; i++) {
    LayerInfo * l = layers.arr[i];
    if (l != NULL) {
      if (l->me_in_layer == true) {

      // Forward to this layer only if predecessor not in this layer
        if (l->ag_list.Locate(pred_aid) == NULL) { 
	  count_pkts += send_data_pkt_to_layer(l,src_aid,src_nid,src_seqno);
        }
      }
    }
  }

  if (state == JOIN) {
    if ( (curr_join_q_lid < MAX_LAYERS) && (curr_join_q_lid > 0) ) {

      LayerInfo * q_l = layers.arr[curr_join_q_lid];
      assert (q_l != NULL);
      assert (q_l->me_in_layer == false);
      count_pkts += send_data_pkt_to_layer(q_l,src_aid,src_nid,src_seqno);
    }
    if (curr_join_q_lid < 0) { // Starting join query at BSE

      // For join to higher layers, we probably already have the higher
      // target layer from the previous leader which left the cluster or
      // did a root xfer to me. Can use that as a fall back option.
      // This should not be for layer 0.
      LayerInfo * target_layer = layers.arr[target_join_q_lid];
      if (target_layer != NULL) {
	assert (target_layer->me_in_layer == false);
	count_pkts += send_data_pkt_to_layer(target_layer,src_aid,src_nid,src_seqno);
      }
    }
  }

  if (is_source == true) {
       AppPacket *ap = new AppPacket(PACKET_DATA);
       ap->u.data_p.original_src.agent_id = src_aid;
       ap->u.data_p.original_src.node_id = src_nid;
       ap->u.data_p.seq_no = src_seqno;
       ap->u.data_p.lid = -1;
#ifdef LOG_DATA_PACKET_SEND
       printf ("[coop %d %d ] At %8.4f send data-pkt : < [ %d %d ] > seq %d : to < [ %d %d ] > bse\n", id, n->id, Scheduler::Clock(), src_aid, src_nid, src_seqno, bse.agent_id, bse.node_id);
#endif // LOG_DATA_PACKET_SEND
       send_pkt_wrapper(ap,bse.agent_id, bse.node_id);
       count_pkts ++;
  }

#ifdef LOG_DATA_PACKET_SEND_COUNT
  if (count_pkts > 0)
    printf ("[coop %d %d ] At %8.4f fwd-data count %d for < [ %d %d ]> seq-no %d\n",id, n->id, Scheduler::Clock(), count_pkts, src_aid, src_nid, src_seqno);
#endif // LOG_DATA_PACKET_SEND_COUNT
  return;
}

int coopAgent::send_data_pkt_to_layer (LayerInfo * l, int src_aid, int src_nid, int src_seqno) {

  int count_pkts = 0;

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = l->ag_list.GetAt(pos);
    if (self_check(this_la) == false) {
      if ( ( (bse.agent_id != this_la->ag.agent_id) || (bse.node_id != this_la->ag.node_id) ) &&
         ( ( (this_la->ag.agent_id != src_aid) || (this_la->ag.node_id != src_nid) ) ) ) {
       AppPacket *ap = new AppPacket(PACKET_DATA);
       ap->u.data_p.original_src.agent_id = src_aid;
       ap->u.data_p.original_src.node_id = src_nid;
       ap->u.data_p.seq_no = src_seqno;
       ap->u.data_p.lid = l->lid;
#ifdef LOG_DATA_PACKET_SEND
       printf ("[coop %d %d ] At %8.4f send data-pkt : < [ %d %d ] > seq %d : to < [ %d %d ] > lid %d\n", id, n->id, Scheduler::Clock(), src_aid, src_nid, src_seqno, this_la->ag.agent_id, this_la->ag.node_id, l->lid);
#endif // LOG_DATA_PACKET_SEND
       send_pkt_wrapper(ap,this_la->ag.agent_id, this_la->ag.node_id);
       count_pkts ++;
     }
   }
 }

  return count_pkts;
}

/* Have to be the cluster root to accept this merge */
/* If we want to accept cluster merge at non-root agents, then
 * we have to change the mechanism of the merge, i.e. the merging
 * root should be able to first figure out which is the correct
 * root, and then advise the children appropriately of its
 * change to be the new correct root.
 */
bool coopAgent::valid_cluster_merge_packet (AppPacket * ap) {

  LayerInfo *l = layers.arr[ap->u.clustermerge_p.layer_id];
  if (l == NULL)
    return false;
  if (l->me_in_layer == false)
    return false;

  if (self_check(l->root) == false)
    return false;
  return true;
}

bool coopAgent::valid_cluster_refresh_packet (AppPacket * ap) {

  LayerInfo * l = layers.arr[ap->u.clusterrefresh_p.layer_id];
  if (l == NULL)
    return false;
  if (l->me_in_layer == false)
    return false;

  void * pos = l->ag_list.Locate(ap->src_agent);
  if (pos == NULL)
    return false;
  return true;
}

bool coopAgent::match_cluster_to_root_info (LayerInfo *l) {

  bool has_change = false;

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la_ag = l->ag_list.GetAt(pos);
    la_ag->valid_tmp = false;
  }

  for (int i = 0; i < l->root->agent_dist_arr_count; i++) {

    void * pos = l->ag_list.Locate(l->root->agent_dist_arr[i].ag.agent_id);
    LayerAgentInfo * this_la;
    if (pos == NULL) {
      this_la = new LayerAgentInfo (l->root->agent_dist_arr[i].ag.agent_id,l->root->agent_dist_arr[i].ag.node_id);
      assert (self_check(this_la) == false); // Cannot have self unknown
      l->AddClusterMember(this_la);
      has_change = true;
    }
    else {
      this_la = l->ag_list.GetAt(pos);
    }
    this_la->valid_tmp = true;
  }

  // Purge the members that are not in the cluster any more
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       ) {
    
    LayerAgentInfo * la_ag = l->ag_list.GetAt(pos);
    void * old_pos = pos;
    l->ag_list.GetNext(pos);
    if (la_ag->valid_tmp == false) {
      l->ag_list.RemoveAt(old_pos);
      assert (la_ag != NULL);
      delete la_ag;
      has_change = true;
    }
  }

  return has_change;
}

/* Send query to the next lower layer if needed (i.e. to
 * one below this_lid).
 * If this is the target layer, then need to figure out point of
 * attachment
 */
void coopAgent::continue_join_query (int this_lid) {

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f (c-cjq) : continue-join-query : lid %d\n", id, n->id, Scheduler::Clock(), this_lid);
#endif // LOG_COOP_JUNK

  if (this_lid > target_join_q_lid) { /* Probe all the cluster members */
    LayerInfo *l = layers.arr[this_lid];
    void * pos = l->ag_list.Locate(id);
    if (pos != NULL) { // I am considered to be part of this cluster !
      assert (self_check(l->root) == false); // I cannot be the root
      delete_self_from_layer(l);
      }
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-cjq) : join-continue\n", id, n->id);
#endif // LOG_COOP_JUNK

    send_join_query_to_cluster(this_lid);
    m_join_resp_received = 0;
    curr_join_q_lid = this_lid;
    jqt.SetTimer(CONST_JOIN_QUERY_TIMEOUT);
    return;
  }
  else { /* Find point of attachment */
    assert (this_lid == target_join_q_lid);
    LayerInfo * lbase = layers.arr[this_lid];
    LayerAgentInfo *self = new LayerAgentInfo(id,n->id);
    self->dist = 0.0;
    if (lbase->AddClusterMember(self) == false) {
      delete self;
      self = lbase->FindClusterMember(id);
      assert(self->dist >= 0.0);
    }
    lbase->me_in_layer = true;
    
    assert (lbase->root != NULL);
    if (lbase->root->dist < 0.0)
      assert (lbase->root->dist >= 0.0);

    AppPacket *pkt = new AppPacket(JOIN_QUERY);
    init_join_query_packet(pkt,this_lid,Scheduler::Clock(),true);
    send_pkt_wrapper(pkt,lbase->root->ag.agent_id, lbase->root->ag.node_id);
    lbase->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);
    lbase->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);
    state = ATTACHED;

    set_higher_layer_ping_timer();

#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-cjq) : attach to my-root < [ %d %d ] >\n", id, n->id, lbase->root->ag.agent_id,lbase->root->ag.node_id);
#endif // LOG_COOP_JUNK
    log_cluster_change_info(lbase->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
    display_cluster_info(lbase,lbase->lid);
#endif // LOG_COOP_CLUSTER_CHANGE

  }
  return;
}

bool coopAgent::valid_join_query_forward_packet (int qlid, bool attach) {
  
  if ( (qlid < 0) || ( (qlid == 0) && (attach == false) ) ) { 
    /* Wrong packet */
    printf ("[Err] Found packet with qlid %d\n",qlid);
    assert (0);
    return false;
  }  
  int lower_lid;
  if (attach == true) /* This is an attach request */
    lower_lid = qlid;
  else
    lower_lid = qlid - 1;
  if (layers.arr[lower_lid] == NULL)
    return false;
  /* I must be part of this cluster. */
  if (layers.arr[lower_lid]->me_in_layer == false)
    return false;

  return true;
}

/* Returns true if this is a valid join response packet that
 * we are expecting
 */
bool coopAgent::valid_join_response_packet (AppPacket *ap) {

  if (state != JOIN)
    return false;

  int this_lid = ap->u.joinresp_p.layer_id;
  if ( (ap->src_agent == bse.agent_id) && (ap->src == bse.node_id) ) {
    if (curr_join_q_lid != -1) // BSE-JF && (curr_join_q_lid != (this_lid+1) )
      return false;
    if (target_join_q_lid > this_lid) {
      assert (0);
      return false;
    }
  }
  else { /* Check for other layer queries */
    if (curr_join_q_lid != (this_lid+1))
      return false;
    LayerInfo * last_layer = layers.arr[curr_join_q_lid];
    if (last_layer == NULL)
      return false;
    if (last_layer->ag_list.Locate(ap->u.joinresp_p.exp_src.agent_id) == NULL)
      return false;
  }

  return true;
}

/* Sends a join query to all members specified in the given
 * layer. The expected response is all members in the immediate
 * lower layer.
 */
void coopAgent::send_join_query_to_cluster (int layer_lid) {

  LayerInfo * this_layer = layers.arr[layer_lid];
#ifdef LOG_COOP_SEND
    printf ("[coop %d %d ] At %8.4f (c-jqtc) : join-query-to-cluster : lid %d\n", id, n->id, Scheduler::Clock(), layer_lid);
#endif // LOG_COOP_SEND

  for (void * pos = this_layer->ag_list.GetHeadPosition();
       pos != NULL;
       this_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * coop_ag = this_layer->ag_list.GetAt(pos);
    if ( (coop_ag->ag.agent_id == bse.agent_id) && (coop_ag->ag.node_id == bse.node_id) ) // Dont send this join query to the BSE
      continue;
    AppPacket *p = new AppPacket(JOIN_QUERY);
    init_join_query_packet (p,layer_lid,Scheduler::Clock(),false);
    send_pkt_wrapper(p,coop_ag->ag.agent_id,coop_ag->ag.node_id);

#ifdef LOG_COOP_SEND
    printf ("[coop %d %d ] . . (c-jqtc) < [ %d %d ] >\n", id, n->id, coop_ag->ag.agent_id, coop_ag->ag.node_id);
#endif // LOG_COOP_SEND
  }

#ifdef LOG_COOP_SEND
  printf ("\n");
#endif // LOG_COOP_SEND
  return;
}

void coopAgent::cancel_higher_layer_ping_timers (void) {

  if (m_ping_in_progress == true) {
    m_ping_in_progress = false;
    m_hlprt.CancelTimer();
    flush_lower_layers(layers.arr[m_ping_lid]);
  }
  m_hlpit.CancelTimer();
  return;
}

void coopAgent::set_higher_layer_ping_timer (void) {
  m_hlpit.SetTimer(CONST_HIGHER_LAYER_PING_INTERVAL_TIMEOUT);
  return;
}

void coopAgent::handle_join_query_timeout (void) {

  process_received_join_responses();
  return;
}

void coopAgent::process_received_join_responses (void) {

  assert (state == JOIN);

  if (curr_join_q_lid  == -1) { /* BSE defaulted ! */
    printf ("[coop %d %d ] At %8.4f BSE did not respond to join-query\n", id, n->id, Scheduler::Clock());
    init_join_query(-1);
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] At %8.4f process-join-resps : bse defaults target lid %d\n", id, n->id, Scheduler::Clock(), target_join_q_lid);
#endif // LOG_COOP_JUNK
    return;
  }

  /* Find the closest agent to this member and re-start
   * the query business for the lower layer */
  LayerInfo * this_layer = layers.arr[curr_join_q_lid];

  void * pos = this_layer->ag_list.Locate(id);
  assert (pos == NULL);

  LayerAgentInfo * min_dist_la_ag = find_closest_agent_in_layer(this_layer,NULL,true);

  if (min_dist_la_ag != NULL) {
    /* Send next set of queries if needed. */
    // assert (layers.arr[curr_join_q_lid] == NULL); THIS ONE WAS WRONG
    if (min_dist_la_ag->dist < 0.0)
      printf ("[Err] At %f [ %d %d ] found closest agent [ %d %d ] with neg dist\n", Scheduler::Clock(), id, n->id, min_dist_la_ag->ag.agent_id, min_dist_la_ag->ag.node_id);

    int lower_lid = curr_join_q_lid - 1;

    flush_layer(lower_lid);

    layers.arr[lower_lid] = min_dist_la_ag->lower_layer;
    min_dist_la_ag->lower_layer = NULL;
    
    LayerAgentInfo * mdla_in_lower = layers.arr[lower_lid]->FindClusterMember(min_dist_la_ag->ag.agent_id);
    assert (mdla_in_lower != NULL);
    mdla_in_lower->dist = min_dist_la_ag->dist;

#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] At %8.4f process-join-resps : lid %d closest < [ %d %d ] > dist %f\n", id, n->id, Scheduler::Clock(), curr_join_q_lid, min_dist_la_ag->ag.agent_id, min_dist_la_ag->ag.node_id, min_dist_la_ag->dist);
#endif // LOG_COOP_JUNK

    continue_join_query(lower_lid);
  }
  else { /* Have to re-query the higher layer if no one in
	  * curr_join_q_lid responded. */
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] At %8.4f process-join-resps  layer defaults lid %d\n", id, n->id, Scheduler::Clock(), this_layer->lid);
#endif // LOG_COOP_JUNK

    delete this_layer;
    layers.arr[curr_join_q_lid] = NULL;
    int higher_lid = curr_join_q_lid + 1;
    if (higher_lid < MAX_LAYERS) {
      if (layers.arr[higher_lid] != NULL) { /* Retry from the higher layer */
	continue_join_query(higher_lid);
      }
      else { /* Have to re-start with the BSE */
	init_join_query(-1);
      }
    }
  }
  return;
}

/* Find a new leader in the cluster. I am leaving this cluster,
 * so I am not included in this.
 */
LayerAgentInfo * coopAgent::find_new_cluster_leader (LayerInfo *l, bool omit_self) {

  if (l->root != NULL)
    assert (self_check(l->root) == true);

#undef LOG_NEW_LEADER_CHOOSE

#ifdef LOG_NEW_LEADER_CHOOSE
  printf ("[coop %d %d ] : At %8.4f (c-fnl) : find-new-ldr : lid %d  %s\n", id, n->id, Scheduler::Clock(), l->lid, (omit_self == true) ? "omit-self" : "include-self");
#endif // LOG_NEW_LEADER_CHOOSE

  double max_dist = -1.0;
  int max_count = -1;
  LayerAgentInfo * new_leader = NULL;
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = l->ag_list.GetAt(pos);

    if (self_check(this_la) == true) {
      continue;
    }
    // HAVE TO LOOK AT HOW TO ACCOMODATE THIS COUNT THING

    int this_count = count_dist_known_members_for_agent(this_la,omit_self);

#ifdef LOG_NEW_LEADER_CHOOSE
    printf ("[coop %d %d ] : . . (c-fnl) : mbr %d : count %d : ", id, n->id, this_la->ag.agent_id, this_la->agent_dist_arr_count);
    for (int i = 0; i < this_la->agent_dist_arr_count; i++)
      printf ("( %d %f )", this_la->agent_dist_arr[i].ag.agent_id, this_la->agent_dist_arr[i].dist);
    printf ("\n[coop %d %d ] : . . (c-fnl) : this-count %d\n", id, n->id, this_count);
#endif // LOG_NEW_LEADER_CHOOSE

    if (this_count == 0) {
      if (max_count < 0) {
        max_count = 0;
        new_leader = this_la;
      }
      continue;
    }

    assert (this_count > 0);
    if (this_count < max_count)
      continue;
    double this_dist = get_farthest_dist_of_member_to_agent(l,this_la,omit_self,true);
#ifdef LOG_NEW_LEADER_CHOOSE
    printf ("[coop %d %d ] : . . (c-fnl) : farthest-dist %f\n", id, n->id, this_dist);
#endif // LOG_NEW_LEADER_CHOOSE

    assert (this_dist >= 0.0);
    if (this_count > max_count) {
      max_count = this_count;
      max_dist = this_dist;
      new_leader = this_la;
    }
    else { //(this_count == max_count)
      assert (max_dist >= 0.0);
      long this_dist_long = USECS(this_dist);
      long max_dist_long = USECS(max_dist);
      if (this_dist_long < max_dist_long) {
	max_dist = this_dist;
	new_leader = this_la;
      }
    }
  }


  if (omit_self == false) { // Need to compare self too
    void * pos = l->ag_list.Locate(id);
    if (pos != NULL) {

      LayerAgentInfo * self_ag = l->ag_list.GetAt(pos);
      LayerAgentInfo * my_farthest = find_farthest_agent_to_me_in_layer(l);
      int my_count = count_dist_known_members_for_me(l);

#ifdef LOG_NEW_LEADER_CHOOSE
      printf ("[coop %d %d ] : . . (c-fnl) : mbr self : count %d : ", id, n->id, l->ag_list.GetSize() );
      for (void * pos2 = l->ag_list.GetHeadPosition();
	   pos2 != NULL;
	   l->ag_list.GetNext(pos2) ) {
	LayerAgentInfo * la2 = l->ag_list.GetAt(pos2);
	printf ("( %d %f )", la2->ag.agent_id, la2->dist);
      }
      printf ("\n[coop %d %d ] : . . (c-fnl) : my-count %d : farthest ( %d %f )\n", id, n->id, my_count, my_farthest->ag.agent_id, my_farthest->dist);
#endif // LOG_NEW_LEADER_CHOOSE
      
      assert (my_farthest != NULL);
      long max_dist_long = USECS(max_dist);
      long my_farthest_dist_long = USECS(my_farthest->dist);
      if ( (my_count > max_count) ||
	   ( (my_count == max_count) &&
	     ( (max_dist < 0.0) || (my_farthest_dist_long <= max_dist_long) ) ) ) {
	// The my_farthest->dist <= max_dist ensures that the current incumbent
	// (i.e. if I am the ldr) continues as ldr if possible
	new_leader = self_ag;
	max_count = my_count;
	max_dist = my_farthest->dist;
      }
    }
  }

#ifdef LOG_NEW_LEADER_CHOOSE
  if (new_leader != NULL)
      printf ("\n[coop %d %d ] : . . (c-fnl) : new-leader %d\n", id, n->id, new_leader->ag.agent_id);
  else
    printf ("\n[coop %d %d ] : . . (c-fnl) : new-leader NULL\n", id, n->id);
#endif // LOG_NEW_LEADER_CHOOSE
  return new_leader;
}

/* Check all the cluster members, if any of them should be the new
 * cluster leader
 */
LayerAgentInfo * coopAgent::check_all_change_cluster_leader (LayerInfo *l) {

  assert (l != NULL);
  // Should be called for a cluster root
  assert ( (l->root->ag.agent_id == id) && (l->root->ag.node_id == n->id) );

  LayerAgentInfo * new_leader = NULL;
  double max_dist = -1.0;
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la_ag = l->ag_list.GetAt(pos);
    if ( (la_ag->ag.agent_id == id) && (la_ag->ag.node_id == n->id) )
      continue;
    double temp_max;
    if (check_change_cluster_leader(l,la_ag,temp_max) == true) {
      assert (temp_max >= 0.0);
      long max_dist_long = USECS(max_dist);
      long temp_max_long = USECS(temp_max);
      if ( (max_dist < 0.0) || (temp_max_long < max_dist_long) ) {
	new_leader = la_ag;
	max_dist = temp_max;
      }
    }
  }

  return new_leader;
}

bool coopAgent::check_change_cluster_leader (LayerInfo *l, LayerAgentInfo *la_ag, double & la_dist) {

  assert (l != NULL);
  // First should check if this member knows all the agents that I know.
  if (are_agent_members_contained_in_mine(l,la_ag) == false)
    return false;

  LayerAgentInfo *farthest = find_farthest_agent_to_me_in_layer(l);
  assert (farthest != NULL);

  la_dist = get_farthest_dist_of_member_to_agent(l,la_ag,false,true);

  long la_dist_long = USECS(la_dist);
  long farthest_dist_long = USECS(farthest->dist);
  if (la_dist_long < farthest_dist_long)
    return true;
  return false;
}

/* ignore_self: true, if we do not consider the distance to self from
 * this agent in the calculation
 * use_only_members_known_to_self: true, if we only consider in this
 * calculation those members that are known to self, and no other
 * members
 */
double coopAgent::get_farthest_dist_of_member_to_agent (LayerInfo *l, LayerAgentInfo * la, bool ignore_self, bool use_only_members_known_to_self) {

#undef LOG_DIST_INFO

#ifdef LOG_DIST_INFO
  printf ("[coop %d %d ] mbr-dist-info : %d : ", id, n->id, la->ag.agent_id);
#endif // LOG_DIST_INFO

  double max_dist = -1.0;
  for (int i = 0; i < la->agent_dist_arr_count; i++) {

    if (la->agent_dist_arr[i].dist < 0.0)
      continue;
    if ( (ignore_self == true) &&
	 ( (la->agent_dist_arr[i].ag.agent_id == id) && (la->agent_dist_arr[i].ag.node_id == n->id) ) )
      continue;

    if (use_only_members_known_to_self == true) {
      void * pos = l->ag_list.Locate(la->agent_dist_arr[i].ag.agent_id);
      if (pos == NULL)
	continue;
      LayerAgentInfo * this_la = l->ag_list.GetAt(pos);
      if (this_la->dist < 0.0)
	continue;
    }
#ifdef LOG_DIST_INFO
    printf ("( %d %5.3f )", la->agent_dist_arr[i].ag.agent_id,la->agent_dist_arr[i].dist);
#endif // LOG_DIST_INFO

    long la_agent_dist_arr_i_dist_long = USECS(la->agent_dist_arr[i].dist);
    long max_dist_long = USECS(max_dist);
    if ( (max_dist < 0.0) || (la_agent_dist_arr_i_dist_long > max_dist_long) ) {
      max_dist = la->agent_dist_arr[i].dist;
    }
  }
#ifdef LOG_DIST_INFO
  printf ("\n");
#endif // LOG_DIST_INFO

  return max_dist;
}

/* Finds the farthest agent in this cluster from this agent */
LayerAgentInfo * coopAgent::find_farthest_agent_to_me_in_layer (LayerInfo * l) {

  assert (l != NULL);

  LayerAgentInfo * farthest = NULL;
  double max_dist = -1.0;
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la_ag = l->ag_list.GetAt(pos);
    if (la_ag->dist < 0.0)
      continue;
    long max_dist_long = USECS(max_dist);
    long la_ag_dist_long = USECS(la_ag->dist);
    if ( (max_dist < 0.0) || (la_ag_dist_long > max_dist_long) ) {
      max_dist = la_ag->dist;
      farthest = la_ag;
    }
  }
  return farthest;
}

/* Find the closest agent in the specified layer among all
 * the agents which have responded
 */
LayerAgentInfo * coopAgent::find_closest_agent_in_layer (LayerInfo * this_layer, AgentInfo * ignore_agent, bool match_join_resp) {

  assert (this_layer != NULL);
  int count_resps = 0;
  double min_dist = -1.0;
  LayerAgentInfo * min_dist_la_ag = NULL;
  for (void * pos = this_layer->ag_list.GetHeadPosition();
       pos != NULL;
       this_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la_ag = this_layer->ag_list.GetAt(pos);
    if (self_check(la_ag) == true)
      continue;

    if (ignore_agent != NULL) {
      if ( (ignore_agent->agent_id == la_ag->ag.agent_id) && (ignore_agent->node_id == la_ag->ag.node_id)) {
	continue;
      }
    }

    if ( (match_join_resp == false) || (la_ag->lower_layer != NULL) ) { 
      /* For join responses, lower_layer != NULL => responded */
      count_resps ++;
      long min_dist_long = USECS(min_dist);
      long la_ag_dist_long = USECS(la_ag->dist);
      if ( (min_dist < 0.0) || (min_dist_long > la_ag_dist_long) ) {
	min_dist = la_ag->dist;
	min_dist_la_ag = la_ag;
      }
    }
  }
  return min_dist_la_ag;
}

/* Checks the member set known to the specified agent. Returns true
 * if all the members known to this agent is also known to me.
 * Known => distance to the member is also known.
 */
bool coopAgent::are_agent_members_contained_in_mine (LayerInfo *l, LayerAgentInfo * la) {

  for (int i = 0; i < la->agent_dist_arr_count; i++) {

    if (la->agent_dist_arr[i].dist >= 0.0) {
      void * pos = l->ag_list.Locate(la->agent_dist_arr[i].ag.agent_id);
      if (pos == NULL)
	return false;
      LayerAgentInfo * this_la = l->ag_list.GetAt(pos);
      if (this_la->dist < 0.0)
	return false;
    }
  }
  return true;
}

bool coopAgent::are_my_members_contained_in_agent (LayerInfo *l, LayerAgentInfo * la) {

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = l->ag_list.GetAt(pos);
    if (this_la->dist >= 0.0) {
      for (int i = 0; i < la->agent_dist_arr_count; i++) {

	if ( (la->agent_dist_arr[i].ag.agent_id == this_la->ag.agent_id) &&
	     (la->agent_dist_arr[i].ag.node_id == this_la->ag.node_id) )
	  continue;
      }
      return false;
    }
  }
  return true;
}

/* ignore self true means that do not include "this" agent in
 * the count
 */
int coopAgent::count_dist_known_members_for_agent (LayerAgentInfo *la, bool ignore_self) {

  int count = 0;
  for (int i = 0; i < la->agent_dist_arr_count; i++) {

    if ( (ignore_self == true) &&
	 ( (la->agent_dist_arr[i].ag.agent_id == id) && (la->agent_dist_arr[i].ag.node_id == n->id) ) )
      continue;
    if (la->agent_dist_arr[i].dist >= 0.0)
      count ++;
  }
  return count;
}

int coopAgent::count_dist_known_members_for_me (LayerInfo *l) {

  assert (l != NULL);
  int count = 0;
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = l->ag_list.GetAt(pos);
    if (la->dist >= 0.0)
      count ++;
  }
  return count;
}

void coopAgent::flush_layer (int layer_id) {

  LayerInfo * this_layer = layers.arr[layer_id];
  if (this_layer != NULL) {
    delete this_layer;
    layers.arr[layer_id] = NULL;
  }
  return;
/*
  for (void * pos = this_layer->ag_list.GetHeadPosition();
       pos != NULL;
       this_layer->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la_ag = this_layer->ag_list.GetAt(pos);
    if (la_ag->lower_layer != NULL) {
      delete la_ag->lower_layer;
      la_ag->lower_layer = NULL;
    }
  }
  return;
*/
}

void coopAgent::send_root_challenge (LayerInfo *l, LayerAgentInfo *la) {

  send_specific_cluster_refresh_packet(l,la,false,true);
  return;
}

/* This is a root transfer to a member which is not part of this
 * cluster.
 */
void coopAgent::do_cluster_merge_root_transfer (LayerInfo *l, AgentInfo *new_root) {

  assert (l != NULL);
  assert ( (l->root != NULL) && (self_check(l->root) == true) );
  assert ( (new_root->agent_id != id) || (new_root->node_id != n->id) );
  AppPacket *ap = new AppPacket(CLUSTER_MERGE);
  put_layer_cluster_info_into_packet(l,ap);
  send_pkt_wrapper(ap,new_root->agent_id,new_root->node_id);

  LayerAgentInfo * root_info = new LayerAgentInfo(new_root->agent_id,new_root->node_id);
  if (l->AddClusterRoot(root_info) == false) {
    delete root_info;
    root_info = l->FindClusterMember(new_root->agent_id);
  }

  l->cr_msgt.CancelTimer();
  send_all_cluster_refresh_packet(l,true,true,root_info);
  l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);

  return;
}

/* Root transfer to a member already part of the cluster */
void coopAgent::do_internal_root_transfer (LayerInfo *l, LayerAgentInfo *la) {

  assert ( (l->root->ag.agent_id == id) && (l->root->ag.node_id == n->id) );
  l->root = la; // This is the root change
  
  l->cr_msgt.CancelTimer();
  send_all_cluster_refresh_packet(l,true,true,NULL);
  l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);

  delete_self_from_all_higher_layers(l->lid);
  return;
}

/* Timeout to send refresh messages to all members in this layer */
void coopAgent::handle_cluster_refresh_msg_timeout (int lid) {

  LayerInfo *l = layers.arr[lid];

  l->cr_msgt.SetTimer(CONST_CLUSTER_REFRESH_MSG_TIMEOUT);

  assert (l->me_in_layer == true);
  assert (l->ag_list.Locate(id) != NULL);
/*
  bool is_root = false;
  if ( (l->root->ag.agent_id == id) && (l->root->ag.node_id == n->id) ) {
    is_root = true;
    LayerAgentInfo * new_leader = check_all_change_cluster_leader(l);
    if (new_leader != NULL) { // Found a new leader
#ifdef LOG_COOP_JUNK
      printf ("[coop] At %8.4f change-ldr < [ %d %d ] > lid %d to < [ %d %d ] >\n", Scheduler::Clock(), id, n->id, l->lid, new_leader->ag.agent_id, new_leader->ag.node_id);
#endif // LOG_COOP_JUNK

      do_internal_root_transfer(l,new_leader);
      // Have to delete self from all upper layers
      delete_self_from_all_higher_layers(l->lid);
      return;
    }
  }
  send_all_cluster_refresh_packet(l,false,is_root,NULL);
*/

  send_all_cluster_refresh_packet(l,false,self_check(l->root),NULL);
  return;
}

void coopAgent::send_all_cluster_refresh_packet (LayerInfo *l, bool set_root_xfer, bool is_root, LayerAgentInfo * ignore_ag) {

  assert (l != NULL);

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = l->ag_list.GetAt(pos);
    if (self_check(la) == true)
      continue;
    if (ignore_ag != NULL) {
      if ( (ignore_ag->ag.agent_id == la->ag.agent_id) && (ignore_ag->ag.node_id == la->ag.node_id) )
	continue;
    }
    send_specific_cluster_refresh_packet(l,la,set_root_xfer,is_root);
  }
  return;
}

/* Sends a refresh packet to a specific cluster members */
void coopAgent::send_specific_cluster_refresh_packet (LayerInfo *l, LayerAgentInfo * la, bool set_root_xfer, bool is_root) {

#ifdef TESTING_JUNK
  void * pos = l->ag_list.Locate(id);
  if (pos != NULL) {
    LayerAgentInfo * self = l->ag_list.GetAt(pos);
    if (self->dist < 0.0)
      assert (self->dist >= 0.0);
  }
#endif // TESTING_JUNK

#ifdef LOG_COOP_SEND
  printf ("[coop %d %d ] At %8.4f send-refresh : lid %d to < [ %d %d ] > ", id, n->id, Scheduler::Clock(), l->lid, la->ag.agent_id, la->ag.node_id);
  printf ("%s %s\n", (is_root == true) ? "root" : ".", (set_root_xfer == true) ? "*xfer*" : ".");
#endif // LOG_COOP_SEND

  AppPacket *ap = new AppPacket (CLUSTER_REFRESH);
  put_layer_cluster_info_into_packet(l,ap);
  if (set_root_xfer == true)
    ap->u.clusterrefresh_p.root_xfer = true;
  if (is_root == true)
    ap->u.clusterrefresh_p.is_root = true;

  if ( (is_root == true) && (set_root_xfer == false) ) {
    if ( (l->lid +1) < MAX_LAYERS ) {
      LayerInfo *hl = layers.arr[l->lid + 1];
      if (hl != NULL) {
	AgentInfo * self = NULL;
	if (set_root_xfer == true) {
	  self = new AgentInfo(id,n->id);
	}
	put_hl_cluster_info_into_packet(hl,ap,self);
	if (self != NULL)
	  delete self;
	// if (hl->me_in_layer == true)
	//  put_hl_cluster_info_into_packet(hl,ap);
	// else
	//  assert (state == JOIN);
      }
    }
  }

  send_pkt_wrapper(ap,la->ag.agent_id,la->ag.node_id);
  return;
}

void coopAgent::handle_cluster_refresh_check_timeout (int lid) {

  LayerInfo *l = layers.arr[lid];
  assert (l != NULL);
  assert (l->me_in_layer == true);

#ifdef LOG_COOP_JUNK
  printf ("\n[coop %d %d ] At %8.4f (c-rc) : refresh-chk : lid %d\n", id, n->id, Scheduler::Clock(), lid);
#endif // LOG_COOP_JUNK

  // If cluster root is lost, then we need to choose a new root.
  // Otherwise do nothing.
  assert (l->root != NULL);
  bool i_am_root = self_check(l->root);

#ifdef LOG_COOP_CLUSTER_PERIODIC
  display_cluster_info(l,lid);
#endif // LOG_COOP_CLUSTER_PERIODIC

  if ( i_am_root == false ) {
    if (l->root->refresh == false) { // Root is gone !
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-rc) : root-lost < [ %d %d ] >\n", id, n->id, l->root->ag.agent_id, l->root->ag.node_id);
#endif // LOG_COOP_JUNK

      l->root = NULL;
    }
  }

  bool change = false;

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       ) {

    LayerAgentInfo * la_ag = l->ag_list.GetAt(pos);
    void * old_pos = pos;
    l->ag_list.GetNext(pos);
    if (self_check(la_ag) == true)
      continue;
    if (la_ag->refresh == false) {
      /* Member is lost */
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-rc) : < [ %d %d ] > *lost*\n", id, n->id, la_ag->ag.agent_id, la_ag->ag.node_id);
#endif // LOG_COOP_JUNK

      l->ag_list.RemoveAt(old_pos);
      delete la_ag;
      change = true;
    }
    else {
      la_ag->refresh = false; // Resetting the refresh flag
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-rc) : < [ %d %d ] >\n", id, n->id, la_ag->ag.agent_id, la_ag->ag.node_id);
#endif // LOG_COOP_JUNK
    }
  }
#ifdef LOG_COOP_JUNK
  printf ("\n");
#endif // LOG_COOP_JUNK

  if (i_am_root == true) {

    if (l->delay_root_xfer == false) { // Cannot do any root transfer
      LayerAgentInfo * new_leader = find_new_cluster_leader(l,false);
      assert (new_leader != NULL);
      
      if (self_check(new_leader) == false) { // Root changes
#ifdef LOG_COOP_JUNK
	printf ("[coop %d %d ] At %8.4f refresh-chk change-ldr : lid %d to < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), l->lid, new_leader->ag.agent_id, new_leader->ag.node_id);
#endif // LOG_COOP_JUNK      
	do_internal_root_transfer(l,new_leader);
	change = true;
      }
      else { // I am still root
	if (l->ag_list.GetSize() < LOWER_K) {
#ifdef LOG_COOP_JUNK
	  printf ("[coop %d %d ] At %8.4f refresh-chk merge : lid %d\n", id, n->id, Scheduler::Clock(),lid);
#endif // LOG_COOP_JUNK
	  
	  if (do_merge_to_peer(l) == true)
	    change = true;
	}
	else if (l->ag_list.GetSize() >= UPPER_3K) {
#ifdef LOG_COOP_JUNK
	  printf ("[coop %d %d ] At %8.4f refresh-chk split : lid %d\n", id, n->id, Scheduler::Clock(), lid);
#endif // LOG_COOP_JUNK
	  
	  split_cluster_using_two_partition(l);
	  change = true;
	}
      }
    }
    else {
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] At %8.4f refresh-chk delay-root-xfer-chk lid %d\n", id, n->id, Scheduler::Clock(), lid);
#endif // LOG_COOP_JUNK
      l->delay_root_xfer = false;
    }
  }
  else { // I am not root
    if (l->root == NULL) { // i.e. root has dissappeared
      assert (change == true);
      LayerAgentInfo * new_leader = find_new_cluster_leader(l,false);
      assert (new_leader != NULL);
      l->root = new_leader;

      cancel_higher_layer_ping_timers();

      if (self_check(new_leader) == true) { // Now I am root
#ifdef LOG_COOP_JUNK
	printf ("[coop %d %d ] At %8.4f refresh-chk me-new-ldr : lid %d\n", id, n->id, Scheduler::Clock(), lid);
#endif // LOG_COOP_JUNK
	l->delay_root_xfer = true;
	join_to_next_higher_layer(l->lid); // Join next higher layer
      }
      else {
	set_higher_layer_ping_timer();
      }
    }
  }
  l->cr_chkt.SetTimer(CONST_CLUSTER_REFRESH_CHECK_TIMEOUT);

  if (change == true) {
    log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
    display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
  }

  return;
}

int coopAgent::get_ping_response_count_in_layer (LayerInfo * l) {

  int count = 0;

  assert (l != NULL);
  assert (m_ping_in_progress == true);

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    if ((l->ag_list.GetAt(pos))->lower_layer != NULL)
      count ++;
  }
  return count;
}

void coopAgent::handle_higher_layer_ping_response_timeout (void) {

  int lower_lid = m_ping_lid - 1;
  if (process_ping_responses() == true) {
    //printf ("[%d %d] At %8.4f ping hl-ping-timeout\n", id, n->id, Scheduler::Clock());
    LayerInfo *l = layers.arr[lower_lid];
    assert (l != NULL);
    log_cluster_change_info(l->lid);
#ifdef LOG_COOP_CLUSTER_CHANGE
    display_cluster_info(l,l->lid);
#endif // LOG_COOP_CLUSTER_CHANGE
  }
  return;
}

void coopAgent::handle_higher_layer_ping_interval_timeout (void) {

  assert (state == ATTACHED);
  LayerInfo *l = highest_layer_self_present();
  assert (l != NULL);
  if ( (l->root->ag.agent_id != bse.agent_id) || (l->root->ag.node_id != bse.node_id) ) {
    if ( (l->lid + 1) < MAX_LAYERS) {
      LayerInfo * next_l = layers.arr[l->lid+1];
      if (next_l != NULL) {
	if (next_l->ag_list.Locate(l->root->ag.agent_id) != NULL) {
	  // My highest layer root should be present in the next higher layer
	  m_ping_in_progress = true;
	  m_ping_lid = next_l->lid;
	  m_ping_resps_received = 0;
	  flush_lower_layers (next_l);
	  do_ping_higher_layer(next_l);
	}
      }
    }
  }

  // Not needed: Done in process_ping_responses()
  // m_hlpit.SetTimer(CONST_HIGHER_LAYER_PING_INTERVAL_TIMEOUT);
  return;
}

void coopAgent::flush_lower_layers (LayerInfo *l) {

  assert (l != NULL);

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = l->ag_list.GetAt(pos);
    if (la->lower_layer != NULL) {
      delete la->lower_layer;
      la->lower_layer = NULL;
    }
  }
  return;
}

LayerInfo * coopAgent::highest_layer_self_present (void) {

  for (int i = 0; i < MAX_LAYERS; i++) {
    LayerInfo *l = layers.arr[i];
    if (l == NULL) {
      assert (i >= 1);
      return layers.arr[i-1];
    }
    if (l->me_in_layer == false) {
      assert (i >= 1);
      return layers.arr[i-1];
    }
  }

  return layers.arr[MAX_LAYERS - 1];
}

void coopAgent::do_ping_higher_layer (LayerInfo *l) {

  assert (l->lid > 0);
  LayerInfo * lower = layers.arr[l->lid - 1];
  assert (lower != NULL);
  assert (lower->root != NULL);

  bool ping_sent = false;

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * this_la = l->ag_list.GetAt(pos);
    if ( ( (lower->root->ag.agent_id == this_la->ag.agent_id) && (lower->root->ag.node_id == this_la->ag.node_id) ) ||
	 ( (this_la->ag.agent_id == bse.agent_id) && (this_la->ag.node_id == bse.node_id) )
	 )
      continue;
#ifdef LOG_COOP_JUNK_PING
    printf ("[coop %d %d ] At %8.4f ping : lid %d : < [ %d %d ] >\n", id, n->id, Scheduler::Clock(), l->lid, this_la->ag.agent_id, this_la->ag.node_id);
#endif // LOG_COOP_JUNK_PING
    AppPacket *ap = new AppPacket (PING_QUERY);
    ap->u.pingq_p.lid = l->lid;
    ap->u.pingq_p.src_time = Scheduler::Clock();
    send_pkt_wrapper(ap, this_la->ag.agent_id, this_la->ag.node_id);
    ping_sent = true;
  }

  if (ping_sent == true)
    m_hlprt.SetTimer(CONST_HIGHER_LAYER_PING_RESPONSE_TIMEOUT);
  else {
    m_ping_in_progress = false;
    m_hlpit.SetTimer(CONST_HIGHER_LAYER_PING_INTERVAL_TIMEOUT);
  }
  return;
}

/* Returns true if this is successfully possible */
bool coopAgent::do_merge_to_peer (LayerInfo *l) {

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f (c-mtp) : merge-to-peer : lid %d\n", id, n->id, Scheduler::Clock(), l->lid);
#endif // LOG_COOP_JUNK

  bool merge_to_bse = false;
  LayerAgentInfo * closest = NULL;
  if ( (l->lid + 1) == MAX_LAYERS)
    merge_to_bse = true;
  else {
    LayerInfo * next_layer = layers.arr[l->lid+1];
    if (next_layer == NULL) {
      assert (state == JOIN);
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-mtp) : higher layer unknown\n", id, n->id);
#endif // LOG_COOP_JUNK
      return false; // Nothing to do till I have the next layer
    }
    void * pos = next_layer->ag_list.Locate(id);
    if (pos == NULL) {
      assert (state == JOIN);
#ifdef LOG_COOP_JUNK
      printf ("[coop %d %d ] . . (c-mtp) : higher layer unknown\n", id, n->id);
#endif // LOG_COOP_JUNK
      return false; // Nothing to do ....
    }
    AgentInfo * ignore_bse = NULL;
    if ( (next_layer->root->ag.agent_id == bse.agent_id) && (next_layer->root->ag.node_id == bse.node_id) ) { // Next layer root is BSE
      if (next_layer->ag_list.GetSize() == 2) { // i.e. me and BSE only
	merge_to_bse = true;
      }
      else {
	ignore_bse = &bse;
      }
    }
    
    closest = find_closest_agent_in_layer(next_layer,ignore_bse,false);
  }

  AgentInfo merge_to;
  if (merge_to_bse == true) {
    merge_to.agent_id = bse.agent_id;
    merge_to.node_id = bse.node_id;
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-mtp) : merging to bse\n", id, n->id);
#endif // LOG_COOP_JUNK
  }
  else if (closest != NULL) {
    merge_to.agent_id = closest->ag.agent_id;
    merge_to.node_id = closest->ag.node_id;
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] . . (c-mtp) : merging to < [ %d %d ] > lid %d\n", id, n->id, closest->ag.agent_id, closest->ag.node_id, l->lid);
#endif // LOG_COOP_JUNK
  }
  else {
    return false;
  }

  do_cluster_merge_root_transfer(l,&merge_to);
  delete_self_from_all_higher_layers(l->lid);
  return true;
}

void coopAgent::join_to_next_higher_layer (int lid) {

  assert ( (lid + 1) < MAX_LAYERS);
  LayerInfo *next_layer = layers.arr[lid+1];

  assert (m_ping_in_progress == false);

  bool join = false;
  if (next_layer == NULL)
    join = true;
  else if (next_layer->me_in_layer == false)
    join = true;

  if (join == true) {
    init_join_query(lid+1);
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] At %8.4f join-to-next-layer : lid %d\n", id, n->id, Scheduler::Clock(), lid+1);
#endif // LOG_COOP_JUNK
  }
  else {
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] At %8.4f join-to-next-layer : lid %d *not needed*\n", id, n->id, Scheduler::Clock(), lid+1);
#endif // LOG_COOP_JUNK
  }
  return;
}

void coopAgent::delete_self_from_all_higher_layers (int lid) {

  cancel_higher_layer_ping_timers();

  if (state == JOIN) {
    if (target_join_q_lid >= (lid+1)) {
#ifdef LOG_COOP_JUNK
    printf ("[coop %d %d ] At %8.4f delete join to : lid %d\n", id, n->id, Scheduler::Clock(), target_join_q_lid);
#endif // LOG_COOP_JUNK
      target_join_q_lid = -1;
      state = ATTACHED;
      jqt.CancelTimer();
    }
  }
  
  for (int i = MAX_LAYERS-1; i > lid; i--) {
    if (layers.arr[i] != NULL) {
      if (layers.arr[i]->me_in_layer == true) {
	delete_self_from_layer(layers.arr[i]);
	// delete layers.arr[i]; // Delete the cluster too
	// layers.arr[i] = NULL;
	layers.arr[i]->cr_msgt.CancelTimer();
	layers.arr[i]->cr_chkt.CancelTimer();
        log_cluster_change_info(i);
#ifdef LOG_COOP_CLUSTER_CHANGE
	display_cluster_info(NULL,i);
#endif // LOG_COOP_CLUSTER_CHANGE
      }
    }
  }
  return;
}

void coopAgent::delete_self_from_layer (LayerInfo *l) {

  assert (l != NULL);
  assert (l->ag_list.Locate(id) != NULL);

  bool i_was_root = self_check(l->root);

#ifdef LOG_COOP_JUNK
  printf ("[coop %d %d ] At %8.4f delete-self-from-layer : lid %d ", id, n->id, Scheduler::Clock(), l->lid);
  if (i_was_root == true)
    printf ("*me root*\n");
  else
    printf ("\n");
#endif // LOG_COOP_JUNK

  bool root_xfer = false;
  if ( i_was_root == true) {
    // I am root, have to find a new root
    root_xfer = true;
    LayerAgentInfo *new_leader = NULL;
    if (l->ag_list.GetSize() > 1) { // If its == 1, then I am the only person
      new_leader = find_new_cluster_leader(l,true);
      assert (new_leader != NULL);
    }
    if (new_leader == NULL) {
      // This implies that there is no other member in this layer. 
    }
    l->root = new_leader;
  }
  // Delete self
  void * pos = l->ag_list.Locate(id);
  assert (pos != NULL);
  LayerAgentInfo * self_ag = l->ag_list.GetAt(pos);
  delete self_ag;
  l->ag_list.RemoveAt(pos);
  l->me_in_layer = false;
  
  if (l->ag_list.GetSize() > 0)
    send_cluster_remove(l,root_xfer);
  return;
}

void coopAgent::send_cluster_remove (LayerInfo *l, bool root_xfer) {

#ifdef LOG_COOP_SEND
  printf ("[coop %d %d ] At %8.4f (c-scx) self-cluster-remove : lid %d : new-root < [ %d %d ] >\n",  id, n->id, Scheduler::Clock(), l->lid, l->root->ag.agent_id, l->root->ag.node_id);
#endif // LOG_COOP_SEND

  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {

    LayerAgentInfo * la = l->ag_list.GetAt(pos);
#ifdef LOG_COOP_SEND
      printf ("[coop %d %d ] . . (c-scx) : < [ %d %d ] >\n", id, n->id, la->ag.agent_id, la->ag.node_id);
#endif // LOG_COOP_SEND

    assert (self_check(la) == false);
    AppPacket *ap = new AppPacket (CLUSTER_REFRESH);
    LayerInfo * hl = NULL;
    if ( ( (l->lid+1) < MAX_LAYERS) && (root_xfer == true) )
      hl = layers.arr[l->lid + 1];
    put_cluster_remove_info_into_packet(l,ap,root_xfer,hl,id,n->id);
    send_pkt_wrapper(ap,la->ag.agent_id,la->ag.node_id);
  }
#ifdef LOG_COOP_SEND
  printf ("\n");
#endif // LOG_COOP_SEND
  return;
}

void coopAgent::log_cluster_change_info (int lid) {

  m_last_change[lid] = Scheduler::Clock();
  return;
}

void coopAgent::display_cluster_info (LayerInfo *l, int lid) {

  if (l == NULL) {
    printf ("[coop %d %d ] At %8.4f cluster-info : lid %d : deleted\n", id, n->id, Scheduler::Clock(), lid);
    return;
  }

  printf ("[coop %d %d ] At %8.4f cluster-info : lid %d : ldr %s %d : count %d : ", id, n->id, Scheduler::Clock(), lid, (self_check(l->root) == true) ? "self" : "other", l->root->ag.agent_id, l->ag_list.GetSize() );

#ifdef LOG_CLUSTER_INFO_DETAILED
  for (void * pos = l->ag_list.GetHeadPosition();
       pos != NULL;
       l->ag_list.GetNext(pos) ) {
    printf (" %d", (l->ag_list.GetAt(pos))->ag.agent_id);
  }
#endif // LOG_CLUSTER_INFO_DETAILED
  printf ("\n");
  return;
}

void coopAgent::init_packet_cache (void) {

  for (int i = 0; i < PACKET_CACHE_SIZE; i++)
    m_pkt_cache_array[i].valid = false;

  return;
}

bool coopAgent::add_to_packet_cache (int src_aid, int src_nid, int seq_no) {

  for (int i = 0; i < PACKET_CACHE_SIZE; i++) {
    if ( (m_pkt_cache_array[i].valid == true) &&
   	 (m_pkt_cache_array[i].ag.agent_id == src_aid) &&
	 (m_pkt_cache_array[i].ag.node_id == src_nid) &&
	 (m_pkt_cache_array[i].seq_no == seq_no) ) {
      return false;
    }
  }

  for (int i = 1; i < PACKET_CACHE_SIZE; i++) {
    m_pkt_cache_array[i-1].valid = m_pkt_cache_array[i].valid;
    m_pkt_cache_array[i-1].ag.agent_id = m_pkt_cache_array[i].ag.agent_id;
    m_pkt_cache_array[i-1].ag.node_id = m_pkt_cache_array[i].ag.node_id;
    m_pkt_cache_array[i-1].seq_no = m_pkt_cache_array[i].seq_no;
  }
  m_pkt_cache_array[PACKET_CACHE_SIZE-1].valid = true;
  m_pkt_cache_array[PACKET_CACHE_SIZE-1].ag.agent_id = src_aid;
  m_pkt_cache_array[PACKET_CACHE_SIZE-1].ag.node_id = src_nid;
  m_pkt_cache_array[PACKET_CACHE_SIZE-1].seq_no = seq_no;

  return true;
}
 
bool coopAgent::self_check (LayerAgentInfo * la) {

  if ( (la->ag.agent_id == id) && (la->ag.node_id == n->id) )
    return true;
  else
    return false;
}

void JoinQueryTimer::EventHandler (void) {

  ca->handle_join_query_timeout();
  return;
}

void HigherLayerPingIntervalTimer::EventHandler (void) {

  ca->handle_higher_layer_ping_interval_timeout();
  return;
}

void HigherLayerPingResponseTimer::EventHandler (void) {

  ca->handle_higher_layer_ping_response_timeout();
  return;
}
